# SPDX-License-Identifier: Apache-2.0

"""
Common shared types
"""

# Standard
from typing import Literal, Self, TypeAlias

# Third Party
from pydantic import Field
from typing_extensions import Any
import pydantic


class NoDefaultsMixin:
    """
    Mixin so that we don't need to copy and paste the code to avoid filling JSON values
    with a full catalog of the default values of rarely-used fields.
    """

    @pydantic.model_serializer(mode="wrap")
    def _workaround_for_design_flaw_in_pydantic(self, nxt):
        """
        Workaround for a design flaw in Pydantic that forces users to accept
        unnecessary garbage in their serialized JSON data or to override
        poorly-documented serialization hooks repeatedly.  Automates overriding said
        poorly-documented serialization hooks for a single dataclass.

        See https://github.com/pydantic/pydantic/issues/4554 for the relevant dismissive
        comment from the devs. This comment suggests overriding :func:`dict()`, but that
        method was disabled a year later. Now you need to add a custom serializer method
        with a ``@model_serializer`` decorator.

        See the docs at
        https://docs.pydantic.dev/latest/api/functional_serializers/
        for some dubious information on how this API works.
        See comments below for important gotchas that aren't in the documentation.
        """
        # Start with the value that self.model_dump() would return without this mixin.
        # Otherwise serialization of sub-records will be inconsistent.
        serialized_value = nxt(self)

        # Figure out which fields are set. Pydantic does not make this easy.
        # Start with fields that are set in __init__() or in the JSON parser.
        fields_to_retain_set = self.model_fields_set

        # Add in fields that were set during validation and extra fields added by
        # setattr().  These fields all go to self.model.extra
        if self.model_extra is not None:  # model_extra is sometimes None. Not sure why.
            # model_extra is a dictionary. There is no self.model_extra_fields_set.
            fields_to_retain_set |= set(list(self.model_extra))

        # Use a subclass hook for the additional fields that fall through the cracks.
        fields_to_retain_set |= set(self._keep_these_fields())

        # Avoid changing Pydantic's field order or downstream code that computes a
        # diff over JSON strings will break.
        fields_to_retain = [k for k in serialized_value if k in fields_to_retain_set]

        # Fields that weren't in the original serialized value should be in a consistent
        # order to ensure consistent serialized output.
        # Use alphabetical order for now and hope for the best.
        fields_to_retain.extend(sorted(fields_to_retain_set - self.model_fields_set))

        result = {}
        for f in fields_to_retain:
            if f in serialized_value:
                result[f] = serialized_value[f]
            else:
                # Sometimes Pydantic adds fields to self.model_fields_set without adding
                # them to the output of self.model_dump()
                result[f] = getattr(self, f)
        for f in self._reserialize_these_fields():
            # Sometimes Pydantic's serializer fails to serialize sub-objects correctly
            field_value = getattr(self, f)
            if field_value is not None:
                result[f] = field_value.model_dump()
        return result

    def _keep_these_fields(self) -> tuple[str]:
        """
        Dataclasses that include this mixin can override this method to add specific
        default values to serialized JSON.

        This is necessary for round-tripping to JSON when there are fields that
        determine which dataclass to use for deserialization.
        """
        return ()

    def _reserialize_these_fields(self) -> tuple[str]:
        """
        Dataclasses that include this mixin can override this method to trigger
        replacing the serialized values of fields with the results of calling
        :func:`model_dump()` on this fields.

        This is necessary because Pydantic's serializer sometimes produces incorrect
        outputs for child objects for reasons unknown when called on the parent object.
        """
        return ()


class _ChatMessageBase(pydantic.BaseModel, NoDefaultsMixin):
    """Base class for all message types.

    Due to the vagaries of Pydantic's JSON parser, we use this class only for common
    functionality, and NOT for defining a common dataclass base type. Use the
    :class:`ChatMessage` type alias to annotate a field or argument as accepting all
    subclasses of this one."""

    content: str
    """Every message has raw string content, even if it also contains parsed structured
    content such as a JSON record."""

    def _keep_these_fields(self):
        return ("role",)


class UserMessage(_ChatMessageBase):
    """User message for an IBM Granite model chat completion request."""

    role: Literal["user"] = "user"


class ToolCall(pydantic.BaseModel, NoDefaultsMixin):
    """Format of an entry in the ``tool_calls`` list of an assistant message"""

    id: str | None = None
    name: str

    # This field should adhere to the argument schema from the  associated
    # FunctionDefinition in the generation request that produced it.
    arguments: dict[str, Any] | None


class AssistantMessage(_ChatMessageBase):
    """
    Lowest-common-denominator assistant message for an IBM Granite model chat
    completion request.
    """

    role: Literal["assistant"] = "assistant"
    tool_calls: list[ToolCall] | None = None


class ToolResultMessage(_ChatMessageBase):
    """
    Message containing the result of a tool call in an IBM Granite model chat completion
    request.
    """

    role: Literal["tool"] = "tool"
    tool_call_id: str


class SystemMessage(_ChatMessageBase):
    """System message for an IBM Granite model chat completion request."""

    role: Literal["system"] = "system"


ChatMessage: TypeAlias = (
    UserMessage | AssistantMessage | ToolResultMessage | SystemMessage
)
"""Type alias for all message types. We use this Union instead of the actual base class
:class:`_ChatMessageBase` so that Pydantic can parse the message list from JSON."""


class ToolDefinition(pydantic.BaseModel, NoDefaultsMixin):
    """
    An entry in the ``tools`` list in an IBM Granite model chat completion request.
    """

    name: str
    description: str | None = None

    # This field holds a JSON schema for a record, but the `jsonschema` package doesn't
    # define an object type for such a schema, instead using a dictionary.
    parameters: dict[str, Any] | None = None


class Document(pydantic.BaseModel, NoDefaultsMixin):
    """RAG documents, which in practice are usually snippets drawn from larger
    documents."""

    text: str
    title: str | None = None

    # vLLM requires document IDs to be strings
    doc_id: str | None = None


class ChatTemplateKwargs(pydantic.BaseModel):
    """
    Values that can appear in the ``chat_template_kwargs`` portion of a valid chat
    completion request for a Granite model.
    """

    model_config = pydantic.ConfigDict(
        # Pass through arbitrary additional keyword arguments for handling by
        # model-specific I/O processors.
        arbitrary_types_allowed=True,
        extra="allow",
    )


class VLLMExtraBody(pydantic.BaseModel, NoDefaultsMixin):
    """
    Elements of `vllm.entrypoints.openai.protocol.ChatCompletionRequest` that
    are not part of OpenAI's protocol and need to be stuffed into the
    "extra_body" parameter of a chat completion request.
    """

    documents: list[Document] | None = Field(
        default=None,
        description=(
            "A list of dicts representing documents that will be accessible to "
            "the model if it is performing RAG (retrieval-augmented generation)."
            " If the template does not support RAG, this argument will have no "
            "effect. We recommend that each document should be a dict containing "
            '"title" and "text" keys.'
        ),
    )
    add_generation_prompt: bool = Field(
        default=True,
        description=(
            "If true, the generation prompt will be added to the chat template. "
            "This is a parameter used by chat template in tokenizer config of the "
            "model."
        ),
    )
    chat_template_kwargs: ChatTemplateKwargs | None = pydantic.Field(
        default=None,
        description=(
            "Additional kwargs to pass to the template renderer. "
            "Will be accessible by the chat template. "
            "Restricted to fields that at least one Granite model "
            "supports."
        ),
    )
    guided_json: str | dict | None = Field(
        default=None,
        description=("If specified, the output will follow the JSON schema."),
    )

    model_config = pydantic.ConfigDict(
        # If an input to this library is an actual vLLM chat completion request, then
        # the request will likely contain additional fields. Pass these fields through
        # when processing with `granite-common`.
        extra="allow",
    )

    def _reserialize_these_fields(self):
        """Hook from NoDefaultsMixin.

        We need to set this because subclasses override chat_template_kwargs with a
        class-specific type.
        """
        return ("chat_template_kwargs",)


class ChatCompletion(pydantic.BaseModel, NoDefaultsMixin):
    """
    Subset of the schema of a chat completion request in vLLM's OpenAI-compatible
    inference API that is exercised by Granite models.

    See the class `vllm.entrypoints.openai.protocol.ChatCompletionRequest` for
    more information.
    """

    messages: list[ChatMessage]
    model: str | None = None
    tools: list[ToolDefinition] | None = None
    extra_body: VLLMExtraBody | None = Field(
        default=None,
        description=("Additional VLLM-specific arguments go here."),
    )

    def _documents(self) -> list[Document] | None:
        """Convenience method for internal code to fetch documents attached to the
        chat completion without having to dig into ``extra_body``."""
        if self.extra_body:
            return self.extra_body.documents
        return None

    def _chat_template_kwargs(self) -> ChatTemplateKwargs | None:
        """Convenience method for internal code to fetch chat template arguments
        without having to dig into ``extra_body``."""
        if self.extra_body:
            return self.extra_body.chat_template_kwargs
        return None

    model_config = pydantic.ConfigDict(
        # If an input to this library is an actual vLLM chat completion request, then
        # the request will likely contain additional fields. Pass these fields through
        # when processing with `granite-common`.
        extra="allow",
    )


class GraniteChatCompletion(ChatCompletion):
    """
    Lowest-common-denominator inputs to a chat completion request for an IBM Granite
    model.
    """

    @pydantic.model_validator(mode="after")
    def _validate_vllm_stuff_in_extra_body(self) -> Self:
        """
        Non-standard VLLM fields should be passed via the ``extra_body`` parameter.
        Make sure the user didn't stuff them into the root, which is currently set up
        to allow arbitrary additional fields.
        """
        model_fields = list(VLLMExtraBody.model_fields.keys())
        for attr_name in model_fields:
            if hasattr(self, attr_name):
                raise ValueError(
                    f"Attempted to pass '{attr_name}' parameter at top "
                    f"level of the chat completion request. "
                    f"Please place this parameter at extra_body."
                    f"{attr_name} for compatibility with the OpenAI "
                    f"Python API."
                )
        return self

    @pydantic.model_validator(mode="after")
    def _validate_documents_at_top_level(self) -> Self:
        """Documents for a Granite model chat completion request should be passed in the
        ``documents`` argument at the top level of the ``extra_body`` portion of the
        request.

        Detect cases where the documents are hanging off of ``chat_template_kwargs``
        and sanitize appropriately.
        """
        if self is None:
            # Weird Pydantic corner case
            return self

        if (
            self.extra_body
            and self.extra_body.chat_template_kwargs
            and hasattr(self.extra_body.chat_template_kwargs, "documents")
        ):
            if self.extra_body.documents is not None:
                raise ValueError(
                    "Conflicting values of documents found in top-level "
                    "'documents' parameter and inside "
                    "'chat_template_kwargs'"
                )
            if not isinstance(self.extra_body.chat_template_kwargs.documents, list):
                raise ValueError(
                    "'documents' parameter inside 'chat_template_kwargs' is not a list"
                )

            # Round-trip through dict so that documents field disappears from JSON
            # representation.
            args = self.extra_body.chat_template_kwargs.model_dump()
            self.documents = [Document.model_validate(d) for d in args["documents"]]
            del args["documents"]
            self.extra_body.chat_template_kwargs = ChatTemplateKwargs.model_validate(
                args
            )

        return self


class Logprob(pydantic.BaseModel, NoDefaultsMixin):
    """
    Subset of the vLLM API passing **prompt** log-probabilities back from vLLM's
    OpenAI-compatible server.

    Note that this is different from the API for token logprobs.

    See the class `vllm.entrypoints.openai.protocol.Logprob` for
    more information.
    """

    logprob: float
    rank: int | None = None
    decoded_token: str | None = None


class ChatCompletionLogProb(pydantic.BaseModel, NoDefaultsMixin):
    """
    Subset of the vLLM API passing **token** log-probabilities back from vLLM's
    OpenAI-compatible server.

    Note that this is different from the API for prompt logprobs.

    See the class `vllm.entrypoints.openai.protocol.ChatCompletionLogProb` for
    more information.
    """

    token: str
    logprob: float = -9999.0
    bytes: list[int] | None = None


class ChatCompletionLogProbsContent(ChatCompletionLogProb):
    """
    Subset of the vLLM API passing token log-probabilities back from vLLM's
    OpenAI-compatible server.

    See the class `vllm.entrypoints.openai.protocol.ChatCompletionLogProbsContent` for
    more information.
    """

    top_logprobs: list[ChatCompletionLogProb] = Field(default_factory=list)


class ChatCompletionLogProbs(pydantic.BaseModel, NoDefaultsMixin):
    """
    Subset of the schema of a token logprobs for a single choice in a chat completion
    result in vLLM's OpenAI-compatible inference API.

    See the class `vllm.entrypoints.openai.protocol.ChatCompletionLogProbs` for
    more information.
    """

    content: list[ChatCompletionLogProbsContent] | None = None


class ChatCompletionResponseChoice(pydantic.BaseModel, NoDefaultsMixin):
    """
    Subset of the schema of a single choice in a chat completion result in vLLM's
    OpenAI-compatible inference API that is exercised by Granite intrinsics.

    See the class `vllm.entrypoints.openai.protocol.ChatCompletionResponseChoice` for
    more information.
    """

    index: int
    message: ChatMessage
    logprobs: ChatCompletionLogProbs | None = None
    # per OpenAI spec this is the default
    finish_reason: str | None = "stop"


class ChatCompletionResponse(pydantic.BaseModel, NoDefaultsMixin):
    """
    Subset of the schema of a chat completion result in vLLM's OpenAI-compatible
    inference API that is exercised by Granite intrinsics.

    See the class `vllm.entrypoints.openai.protocol.ChatCompletionResponse` for
    more information.
    """

    choices: list[ChatCompletionResponseChoice]

    # vLLM-specific fields that are not in OpenAI spec
    prompt_logprobs: list[dict[int, Logprob] | None] | None = None

    model_config = pydantic.ConfigDict(
        # Actual response objects will contain additional fields which this library
        # ignores.
        extra="ignore",
    )



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Classes and functions that implement input and output string processing for the Granite
3.2 family of models.
"""

# Standard
import json
import re

# First Party
from granite_common.base.types import (
    ChatCompletion,
)
from granite_common.granite3.constants import (
    NO_TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART,
    NO_TOOLS_NO_DOCS_NO_THINKING_SYSTEM_MESSAGE_PART,
)
from granite_common.granite3.input import Granite3InputProcessor

# Local
from .constants import (
    ALL_SPECIAL_TOKENS,
    DOCS_AND_CITATIONS_SYSTEM_MESSAGE_PART,
    DOCS_AND_HALLUCINATIONS_SYSTEM_MESSAGE_PART,
    MODEL_NAME,
    NO_TOOLS_AND_NO_DOCS_AND_THINKING_SYSTEM_MESSAGE_PART,
    TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART,
    TOOLS_AND_NO_DOCS_SYSTEM_MESSAGE_PART,
)
from .types import Granite32ChatCompletion


class Granite32InputProcessor(Granite3InputProcessor):
    """
    Input processor for version 3.2 of the main Granite models, all sizes.

    This input processor is based on the Jinja template that was used during
    supervised fine tuning of these models. This template is as follows:
    ```
    {%- if messages[0]['role'] == 'system' %}
        {%- set system_message = messages[0]['content'] %}
        {%- set loop_messages = messages[1:] %}
    {%- else %}
        {%- set system_message = \"Knowledge Cutoff Date: April 2024.\nToday's Date: \"
          + strftime_now('%B %d, %Y') + \".\nYou are Granite, developed by IBM.\" %}
        {%- if tools and documents %}
                {%- set system_message = system_message + \" You are a helpful AI
                assistant with access to the following tools.
                  When a tool is required to answer the user's query, respond with
                  <|tool_call|> followed by a JSON list of tools used. If a tool does
                  not exist in the provided list of tools, notify the user that you do
                  not have the ability to fulfill the request.\n\nWrite the response to
                  the user's input by strictly aligning with the facts in the provided
                  documents. If the information needed to answer the question is not
                  available in the documents, inform the user that the question cannot
                  be answered based on the available data.\" %}
        {%- elif tools %}
                {%- set system_message = system_message + \" You are a helpful AI
                assistant with access to the following tools. When a tool is required to
                answer the user's query, respond with <|tool_call|> followed by a JSON
                list of tools used. If a tool does not exist in the provided list of
                tools, notify the user that you do not have the ability to fulfill the
                request.\" %}
        {%- elif documents %}
                {%- set system_message = system_message + \" Write the response to the
                user's input by strictly aligning with the facts in the provided
                documents. If the information needed to answer the question is not
                available in the documents, inform the user that the question cannot be
                answered based on the available data.\" %}
        {%- elif thinking %}
                {%- set system_message = system_message + \" You are a helpful AI
                assistant.\nRespond to every user query in a comprehensive and detailed
                way. You can write down your thoughts and reasoning process before
                responding. In the thought process, engage in a comprehensive cycle of
                analysis, summarization, exploration, reassessment, reflection,
                backtracing, and iteration to develop well-considered thinking process.
                In the response section, based on various attempts, explorations, and
                reflections from the thoughts section, systematically present the final
                solution that you deem correct. The response should summarize the
                thought process. Write your thoughts after 'Here is my thought process:'
                and write your response after 'Here is my response:' for each user
                query.\" %}
        {%- else %}
                {%- set system_message = system_message + \" You are a helpful AI
                assistant.\" %}
        {%- endif %}
        {%- if 'citations' in controls and documents %}
            {%- set system_message = system_message + '\n\nIn your response, use the
            symbols <co> and </co> to indicate when a fact comes from a document in the
            search result, e.g <co>0</co> for a fact from document 0. Afterwards, list
            all the citations with their corresponding documents in an ordered list.' %}
        {%- endif %}
        {%- if 'hallucinations' in controls and documents %}
            {%- set system_message = system_message + '\n\nFinally, after the response
            is written, include a numbered list of sentences from the response that are
            potentially hallucinated and not based in the documents.' %}
        {%- endif %}
        {%- set loop_messages = messages %}
    {%- endif %}
    {{- '<|start_of_role|>system<|end_of_role|>' + system_message +
        '<|end_of_text|>\n' }}
    {%- if tools %}
        {{- '<|start_of_role|>tools<|end_of_role|>' }}
        {{- tools | tojson(indent=4) }}
        {{- '<|end_of_text|>\n' }}
    {%- endif %}
    {%- if documents %}
        {{- '<|start_of_role|>documents<|end_of_role|>' }}
        {%- for document in documents %}
            {{- 'Document ' + loop.index0 | string + '\n' }}
            {{- document['text'] }}
            {%- if not loop.last %}
                {{- '\n\n'}}
            {%- endif%}
        {%- endfor %}
        {{- '<|end_of_text|>\n' }}
    {%- endif %}
    {%- for message in loop_messages %}
        {{- '<|start_of_role|>' + message['role'] + '<|end_of_role|>' +
        message['content'] + '<|end_of_text|>\n' }}
        {%- if loop.last and add_generation_prompt %}
            {{- '<|start_of_role|>assistant' }}
            {%- if controls %}
                {{- ' ' + controls | tojson()}}
            {%- endif %}
            {{- '<|end_of_role|>' }}
        {%- endif %}
    {%- endfor %}
    ```
    """

    def _build_default_system_message(
        self, chat_completion: Granite32ChatCompletion
    ) -> str:
        """
        :chat_completion: Chat completion request that does not include a custom
            system message.
        :returns: The standard system message portion of the prompt for the request,
            as a string suitable to feed to the model's tokenizer.
        """
        # bool([]) == bool(None) == False
        have_documents = bool(chat_completion._documents())
        have_tools = bool(chat_completion.tools)
        have_thinking = chat_completion.thinking()
        controls = chat_completion.controls()

        # Carefully hew to the policy that the original Jinja template's behavior
        # defines.
        # First, disallow the cases that the authors of the Jinja template did not
        # provide any code to handle.
        if have_thinking and have_documents:
            raise ValueError(
                f"'thinking' flag is set, but documents were provided. "
                f"{MODEL_NAME} only supports the 'thinking' flag when "
                f"documents are not provided."
            )
        if have_thinking and have_tools:
            raise ValueError(
                f"'thinking' flag is set, but tools were provided. "
                f"{MODEL_NAME} only supports the 'thinking' flag when "
                f"tools are not provided."
            )

        # The default system message starts with a header that includes the date and
        # knowledge cutoff.
        system_message = "<|start_of_role|>system<|end_of_role|>"
        system_message += Granite32InputProcessor._make_system_message_start()

        # Add a middle part that varies depending on tools, documents, and citations.
        if have_documents and have_tools:
            system_message += TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART
        elif have_documents:  # and not have_tools
            system_message += NO_TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART
        elif have_tools:  # and not have_documents
            system_message += TOOLS_AND_NO_DOCS_SYSTEM_MESSAGE_PART
        elif have_thinking:  # if not have_documents and not have_tools
            system_message += NO_TOOLS_AND_NO_DOCS_AND_THINKING_SYSTEM_MESSAGE_PART
        else:  # if not inputs.thinking and not have_documents and not have_tools
            system_message += NO_TOOLS_NO_DOCS_NO_THINKING_SYSTEM_MESSAGE_PART

        # Next comes an optional section of instructions for citations.
        if controls.citations:
            if not have_documents:
                # TODO: The template skips the citations instruction in this case.
                # Is this behavior an error? Should we raise an error if the caller
                # sets the citations flag but provides zero documents?
                pass
            else:  # if have_documents
                system_message += DOCS_AND_CITATIONS_SYSTEM_MESSAGE_PART

        # Then comes an optional section of instructions for hallucinations.
        if controls.hallucinations:
            if not have_documents:
                raise ValueError(
                    f"'hallucinations' flag is set, but the model input does not "
                    f"include documents. {MODEL_NAME} only supports the "
                    f"'hallucinations' flag when documents are provided."
                )
            # if have_documents
            system_message += DOCS_AND_HALLUCINATIONS_SYSTEM_MESSAGE_PART

        # Finish with an end of text
        system_message += "<|end_of_text|>\n"

        return system_message

    @classmethod
    def _remove_special_tokens(cls, text: str) -> str:
        """
        Removes any special tokens from the text string.

        :param text: String for removal of special tokens.
        :returns: String with any special tokens removed.
        """

        regex_roles = r"<\|start_of_role\|>.*<\|end_of_role\|>.*<\|end_of_text\|>"
        regex_tool_call = r"<\|tool_call\|>\{.*\}"

        new_text = text
        new_text = re.sub(regex_roles, "", new_text)
        new_text = re.sub(regex_tool_call, "", new_text)

        # Replace any stray special tokens.
        for special_token in ALL_SPECIAL_TOKENS:
            new_text = new_text.replace(special_token, "")
        return new_text

    @classmethod
    def sanitize(cls, chat_completion, parts="all"):
        # Call the parent sanitize function with the specific remove special
        # tokens function for this Granite version.
        return super()._sanitize(chat_completion, cls._remove_special_tokens, parts)

    def transform(
        self, chat_completion: ChatCompletion, add_generation_prompt: bool = True
    ) -> str:
        # Downcast to a Model-specific request type with possible additional fields.
        # This operation also performs additional validation.
        chat_completion = Granite32ChatCompletion.model_validate(
            chat_completion.model_dump()
        )
        controls = chat_completion.controls()
        have_thinking = chat_completion.thinking()

        # Check for a caller-provided system message
        system_message_json, loop_messages = self._split_messages(chat_completion)

        if system_message_json is not None:
            if have_thinking:
                raise ValueError(
                    f"'thinking' flag is set, but the model input includes a custom "
                    f"system message. {MODEL_NAME} only supports the "
                    f"'thinking' flag when the default system message is used."
                )
            if chat_completion._documents():
                raise ValueError(
                    f"The model input includes documents and a custom system message. "
                    f"{MODEL_NAME} only supports the documents list when "
                    f"the default system message is used."
                )
            if controls.citations:
                raise ValueError(
                    f"'citations' flag is set, but the model input includes a custom "
                    f"system message. {MODEL_NAME} only supports the "
                    f"'citations' flag when the default system message is used."
                )
            if controls.hallucinations:
                raise ValueError(
                    f"'hallucinations' flag is set, but the model input includes a "
                    f"custom system message. {MODEL_NAME} only supports "
                    f"the 'hallucinations' flag when the default system message is "
                    f"used."
                )
            system_message = (
                f"<|start_of_role|>system<|end_of_role|>"
                f"{system_message_json.content}<|end_of_text|>\n"
            )
        else:  # if system_message_json is None:
            # No caller-provided system message.
            # Create a default system message according to the rules implied by the
            # tokenizer's Jinja template.
            system_message = self._build_default_system_message(chat_completion)

        if not bool(chat_completion.tools):
            tools_part = ""
        else:
            tools_part = (
                "<|start_of_role|>tools<|end_of_role|>"
                + json.dumps([t.model_dump() for t in chat_completion.tools], indent=4)
                + "<|end_of_text|>\n"
            )

        if not bool(chat_completion._documents()):
            documents_part = ""
        else:
            documents_body = "\n\n".join(
                [
                    f"Document {i}\n{chat_completion._documents()[i].text}"
                    for i in range(len(chat_completion._documents()))
                ]
            )
            documents_part = (
                "<|start_of_role|>documents<|end_of_role|>"
                + documents_body
                + "<|end_of_text|>\n"
            )

        messages_part = "".join(
            [self._message_to_prompt_string(message) for message in loop_messages]
        )

        # Jinja template expects arbitrary JSON, while our dataclass has specific
        # fields for supported controls.
        controls_record = self._build_controls_record(chat_completion)
        controls_str = (
            "" if controls_record is None else " " + json.dumps(controls_record)
        )

        generation_prompt_part = (
            ""
            if not add_generation_prompt
            else f"<|start_of_role|>assistant{controls_str}<|end_of_role|>"
        )

        return (
            system_message
            + tools_part
            + documents_part
            + messages_part
            + generation_prompt_part
        )



# SPDX-License-Identifier: Apache-2.0

# Standard

__doc__ = """
Constants used in code that is specific to the Granite 3 family of models, but not 
specific to a particular point release.
"""


# String that a Granite 3.x model must receive immediately after _SYSTEM_MESSAGE_START
# if there are documents in the current request but there are no tools in the current
# request.
NO_TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART = """\
 Write the response to the user's input by strictly aligning with the facts in the \
provided documents. If the information needed to answer the question is not available \
in the documents, inform the user that the question cannot be answered based on the \
available data."""


# String that a Granite 3.x model must receive immediately after _SYSTEM_MESSAGE_START
# if there are no tools or documents in the current request and the "thinking" flag is
# set to `False`.
NO_TOOLS_NO_DOCS_NO_THINKING_SYSTEM_MESSAGE_PART = """\
 You are a helpful AI assistant."""



# SPDX-License-Identifier: Apache-2.0

"""
Type definitions that are shared within the Granite 3 family of models
"""

# Third Party
import pydantic
import pydantic_core

# First Party
from granite_common.base.types import (
    AssistantMessage,
    ChatTemplateKwargs,
    Document,
    GraniteChatCompletion,
    NoDefaultsMixin,
    SystemMessage,
    UserMessage,
    VLLMExtraBody,
)


class Hallucination(pydantic.BaseModel):
    """Hallucination data as returned by the model output parser"""

    hallucination_id: str
    risk: str
    reasoning: str | None = None
    response_text: str
    response_begin: int
    response_end: int


class Citation(pydantic.BaseModel):
    """Citation data as returned by the model output parser"""

    citation_id: str
    doc_id: str
    context_text: str
    context_begin: int
    context_end: int
    response_text: str
    response_begin: int
    response_end: int


class Granite3Controls(
    pydantic.BaseModel,
):
    """
    Granite 3.x controls record
    """

    citations: bool | None = None
    hallucinations: bool | None = None
    length: str | None = None  # Length output control variable
    originality: str | None = None

    @pydantic.field_validator("length", mode="after")
    @classmethod
    def _validate_length(cls, value: str | None) -> str | None:
        if value is None or value == "short" or value == "long":
            return value
        raise pydantic_core.PydanticCustomError(
            "length field validator",
            'length ({length}) must be "short" or "long" or None',
            {"length": value},
        )

    @pydantic.field_validator("originality", mode="after")
    @classmethod
    def _validate_originality(cls, value: str | None) -> str | None:
        if value is None or value == "extractive" or value == "abstractive":
            return value
        raise pydantic_core.PydanticCustomError(
            "originality field validator",
            'originality ({originality}) must be "extractive" or "abstractive" or None',
            {"originality": value},
        )


class Granite3Kwargs(ChatTemplateKwargs, NoDefaultsMixin):
    controls: Granite3Controls | None = None
    thinking: bool = False


class Granite3ChatCompletion(GraniteChatCompletion):
    """
    Class that represents the inputs that are common to models of the IBM Granite 3.x
    family.
    """

    def controls(self) -> Granite3Controls:
        """
        :returns: An appropriate Granite 3 controls record for the chat completion
        """
        if (
            self.extra_body
            and self.extra_body.chat_template_kwargs
            and self.extra_body.chat_template_kwargs.controls
        ):
            return self.extra_body.chat_template_kwargs.controls
        return Granite3Controls()

    def thinking(self) -> bool:
        """
        :returns: ``True`` if thinking mode is enabled for this request
        """
        return (
            self.extra_body
            and self.extra_body.chat_template_kwargs
            and self.extra_body.chat_template_kwargs.thinking
        )

    @pydantic.field_validator("extra_body")
    @classmethod
    def _validate_chat_template_kwargs(cls, extra_body: VLLMExtraBody) -> VLLMExtraBody:
        """
        Validates kwargs that are specific to Granite 3 chat templates and converts
        the ``chat_template_kwargs`` field to a Granite 3-specific dataclass.

        Other arguments are currently passed through without checking.
        """
        if extra_body.chat_template_kwargs:
            kwargs_dict = extra_body.chat_template_kwargs.model_dump()
            extra_body.chat_template_kwargs = Granite3Kwargs.model_validate(kwargs_dict)
        return extra_body

    @pydantic.field_validator("messages")
    @classmethod
    def _validate_inputs_messages(cls, messages: list) -> list:
        # Make a copy so the validation code below can mutate the messages list but pass
        # through the original value. The caller also might have a pointer to the list.
        original_messages = messages
        messages = messages.copy()

        # There is no supervised fine tuning data for the case of zero messages.
        # Models are not guaranteed to produce a valid response if there are zero
        # messages.
        if len(messages) == 0:
            raise ValueError(
                "No messages. Model behavior for this case is not defined."
            )

        # The first message, and only the first message, may be the system message.
        first_message_is_system_message = isinstance(messages[0], SystemMessage)
        if first_message_is_system_message:
            messages = messages[1:]
            # If there is a system message, there must be at least one more user or
            # assistant message.
            if len(messages) == 0:
                raise ValueError(
                    "Input contains only a system message. Model behavior for this "
                    "case is not defined."
                )

        # The first message that is not a system message must be
        # either a user or assistant message.
        if not isinstance(messages[0], UserMessage | AssistantMessage):
            if first_message_is_system_message:
                raise ValueError(
                    f"First message after system message must be a user or "
                    f"assistant message. Found type {type(messages[0])}"
                )
            raise ValueError(
                f"First message must be a system, user, or assistant "
                f"Found type {type(messages[0])}"
            )

        # Undocumented constraint: All other messages form a conversation that
        # alternates strictly between user and assistant, possibly with tool calls
        # after an assistant turn and before the next user turn.
        # TODO: Validate this invariant.

        # Pydantic will use the value that this validator returns as the value of the
        # messages field. Undo any changes that we made during validation and return
        # the original value.
        return original_messages


class Granite3AssistantMessage(AssistantMessage):
    """
    An assistant message augmented with additional fields that are specific to the
    Granite 3 family of models.
    """

    reasoning_content: str | None = None
    citations: list[Citation] | None = None
    documents: list[Document] | None = None
    hallucinations: list[Hallucination] | None = None
    stop_reason: str | None = None

    raw_content: str | None = pydantic.Field(
        default=None,
        description=(
            "Raw response content without any parsing, for debugging and "
            "re-serialization."
        ),
    )



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Dataclasses that are specific to the Granite 3.2 family of models.
"""

# Third Party
import pydantic

# First Party
from granite_common.base.types import Document, VLLMExtraBody
from granite_common.granite3.types import Granite3ChatCompletion


class Granite32ChatCompletion(Granite3ChatCompletion):
    """
    Class that represents the inputs to a Granite 3.2 model generation call.
    """

    @pydantic.field_validator("extra_body")
    @classmethod
    def _validate_documents(cls, extra_body: VLLMExtraBody) -> VLLMExtraBody:
        """
        Granite 3.2 documents should not have document IDs.
        """
        if extra_body.documents:
            for i, d in enumerate(extra_body.documents):
                if not isinstance(d, Document):
                    raise TypeError(
                        f"Expected Document at position {i} but found "
                        f"{d} of type {type(d)}"
                    )
                if d.doc_id is not None:
                    raise ValueError(
                        f"Document at position {i} contains a `doc_id` "
                        f"field. This field is not allowed for Granite "
                        f"3.2."
                    )
        return extra_body



# SPDX-License-Identifier: Apache-2.0

"""
Tests for the core types
"""

# Third Party
import pytest

# First Party
from granite_common.base import types


def test_tool_call():
    fc = types.ToolCall(id="1", name="test_func", arguments={"arg1": "value"})
    assert fc.id == "1"
    assert fc.name == "test_func"
    assert fc.arguments == {"arg1": "value"}


def test_chat_message_types():
    um = types.UserMessage(content="user content")
    am = types.AssistantMessage(content="assistant content", tool_calls=[])
    trm = types.ToolResultMessage(content="tool result content", tool_call_id="123")
    sm = types.SystemMessage(content="system content")

    assert um.role == "user"
    assert am.role == "assistant"
    assert trm.role == "tool"
    assert sm.role == "system"
    assert um.model_dump()["role"] == "user"
    assert am.model_dump()["role"] == "assistant"
    assert trm.model_dump()["role"] == "tool"
    assert sm.model_dump()["role"] == "system"


def test_tool_definition():
    td = types.ToolDefinition(name="test_func", description="Test function")
    assert td.name == "test_func"
    assert td.description == "Test function"
    assert td.parameters is None

    td.model_dump()


def test_chat_completion():
    cc = types.ChatCompletion(messages=[], tools=[])
    assert len(cc.messages) == 0
    assert len(cc.tools) == 0

    # Only fields set in constructor should pass through to JSON
    assert sorted(cc.model_dump().keys()) == ["messages", "tools"]


def test_granite_chat_completion():
    with pytest.raises(ValueError, match="'documents' parameter at top level"):
        types.GraniteChatCompletion.model_validate(
            {"messages": [], "documents": [], "chat_template_kwargs": {"documents": []}}
        )
    with pytest.raises(ValueError, match="Conflicting values of documents"):
        types.GraniteChatCompletion.model_validate(
            {
                "messages": [],
                "extra_body": {
                    "documents": [],
                    "chat_template_kwargs": {"documents": []},
                },
            }
        )
    with pytest.raises(ValueError, match="not a list"):
        types.GraniteChatCompletion.model_validate(
            {
                "messages": [],
                "extra_body": {"chat_template_kwargs": {"documents": "foo"}},
            }
        )

    # Documents in kwargs should be moved to top-level arg
    docs_in_kwargs = types.GraniteChatCompletion.model_validate(
        {
            "messages": [],
            "extra_body": {"chat_template_kwargs": {"documents": [{"text": "hello"}]}},
        }
    )
    assert docs_in_kwargs.documents is not None
    assert not hasattr(docs_in_kwargs.extra_body.chat_template_kwargs, "documents")
    assert (
        "documents" not in docs_in_kwargs.extra_body.chat_template_kwargs.model_dump()
    )



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Output processing code that is specific to the Granite 3 family of models, but not
specific to a particular point release.
"""
# Standard
from collections.abc import Callable
import copy
import logging
import re
import sys

# First Party
from granite_common.util import find_substring_in_text

# Setup logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(
    logging.Formatter("%(levelname)s %(asctime)s %(message)s", datefmt="%H:%M:%S")
)
logger.addHandler(handler)


def create_dict(input_array: object, **key_attrib_names: str) -> dict:
    """
    Given an array of dicts and the name of attribute(s) within the array, return a
    dict containing the contents of the array indexed by the given attribute(s)
    """
    new_dict = {}

    for item in input_array:
        new_dict_key_val: str = ""
        key_attribs_len = len(key_attrib_names)
        # Key for dictionary will be a combinations of attribute(s)
        # the dictionary that we are trying to index
        for key_attrib in key_attrib_names.values():
            new_dict_key_val += item[key_attrib]
            key_attribs_len -= 1
            if key_attribs_len > 0:
                new_dict_key_val += "-"

        if new_dict_key_val in new_dict:
            logger.warning(
                f"Found duplicate item while creating dictionary: "
                f"'{new_dict[new_dict_key_val]}'."
            )

        new_dict[new_dict_key_val] = item

    return new_dict


def parse_hallucinations_text(hallucinations_text: str) -> list[dict]:
    """
    Given the hallucinations text output by model under the "# Hallucinations:"
    section, extract the hallucinations info as an array of the form:

    [
        {
            "hallucination_id": "Hallucination ID output by model",
            "risk": "Hallucination risk flag",
            "response_text": "Substring of response text for which
                                hallucination risk is computed"
        },
        ...
    ]
    """

    hallucinations = []

    # Find begin spans of all hallucinations
    matches_iter = re.finditer(
        "(\\d+)\\.\\s*Risk (low|high|unanswerable):",
        hallucinations_text,
    )
    matches = []
    for match in matches_iter:
        matches.append({"match_begin": match.start()})

    if len(matches) == 0:
        logger.warning(
            "Failed to extract hallucination info."
            "Expected hallucination info but none found."
        )

    # For each hallucination, extract its components (hallucination ID,
    # risk, response text)
    for i in range(len(matches)):  # pylint: disable=consider-using-enumerate
        cur_match = matches[i]

        # Select text corresponding to hallucination (which is the text from the
        # beginning of the hallucination until the beginning of the next
        # hallucination or the end of the text; whichever comes first)
        if i + 1 < len(matches):
            next_match_begin = matches[i + 1]["match_begin"] - 1
        else:
            next_match_begin = len(hallucinations_text)
        hallucination_str = hallucinations_text[
            cur_match["match_begin"] : next_match_begin
        ]

        # Within the hallucination text, extract the citation components
        # (hallucination ID, risk, response text)
        # Use ?s flag to include newlines in match
        matches_iter = re.finditer(
            "(?s)(\\d+)\\.\\s*Risk (low|high|unanswerable): (.+)$",
            hallucination_str,
        )
        idx = 0
        for match in matches_iter:
            cur_hallucination = {
                "hallucination_id": match.group(1),
                "risk": match.group(2),
                "response_text": match.group(3),
            }
            hallucinations.append(cur_hallucination)

            idx += 1

        if idx == 0:
            logger.warning(
                "Error in finding components of hallucination: "
                "Expected single RegEx match but found none."
            )
        if idx > 1:
            logger.warning(
                "Error in finding components of hallucination: "
                "Expected single RegEx match but found several."
            )

    return hallucinations


def add_hallucination_response_spans(
    hallucination_info: list[dict],
    response_text_without_citations: str,
    remove_citations_from_response_text: Callable,
) -> list[dict]:
    """
    Given the response text (cleaned from citation tags) and a
    parsed hallucinations info of the form:

    [
        {
            "hallucination_id": "Hallucination ID output by model",
            "risk": "Hallucination risk flag",
            "response_text": "Substring of response text for which hallucination
                                risk is computed"
        },
        ...
    ]

    add to each hallucination element in the array the following attributes
    (the "response_text" replaces the attribute of the same name):

        "response_text": "The response text corresponding to the hallucination
                            element cleaned from citation tags"
        "response_begin": "The begin index of "response_text" within the response
                            text (without citation tags)"
        "response_end": "The end index of "response_text" within the response
                            text (without citation tags)"
    """

    augmented_hallucination_info = copy.deepcopy(hallucination_info)

    for hallucination in augmented_hallucination_info:
        hallucination_response_text_without_citations = (
            remove_citations_from_response_text(hallucination["response_text"])
        )
        matches = find_substring_in_text(
            hallucination_response_text_without_citations,
            response_text_without_citations,
        )
        if len(matches) == 0:
            logger.warning(
                "Error in adding the response spans to hallucination: "
                "Hallucination text not found in response"
            )
            # Install placeholder values to avoid breaking downstream code.
            hallucination["response_begin"] = 0
            hallucination["response_end"] = 0
            continue

        if len(matches) > 1:
            logger.warning(
                "Hallucination text found multiple times in "
                "response: Selecting first match"
            )
        hallucination["response_text"] = hallucination_response_text_without_citations
        hallucination["response_begin"] = matches[0]["begin_idx"]
        hallucination["response_end"] = matches[0]["end_idx"]

    return augmented_hallucination_info


def add_citation_context_spans(
    citation_info: list[dict], docs: list[dict]
) -> list[dict]:
    """
    Given a set of docs and an array of citations of the form:

    [
        {
            "citation_id": "Citation ID output by model",
            "doc_id": "ID of doc where the cited text is drawn from",
            "context_text": "The cited text from the context"
        },
        ...
    ]

    add to each citation in the array the following two attributes:

        "context_begin": "The begin index of "context_text" within document with
                            ID doc_id"
        "context_end": "The end index of "context_text" within document with ID doc_id"
    """
    augmented_citation_info = copy.deepcopy(citation_info)
    docs_by_cit_doc_id = create_dict(
        docs, citation_attrib="citation_id", document_attrib="doc_id"
    )
    for citation in augmented_citation_info:
        # Init values in event of error in processing
        citation["context_begin"] = 0
        citation["context_end"] = 0
        try:
            dict_id = citation["citation_id"] + "-" + citation["doc_id"]
            doc = docs_by_cit_doc_id[dict_id]
        except KeyError:
            logger.warning(
                f"Document with id: {dict_id} not found "
                f"when adding citation context spans."
            )
            continue

        matches = find_substring_in_text(citation["context_text"], doc["text"])
        if len(matches) == 0:
            logger.warning(
                "Error in adding the context spans to citation: "
                "Cited text not found in corresponding document."
            )
            continue

        if len(matches) > 1:
            logger.warning(
                "Cited text found multiple times in corresponding "
                "document: Selecting first match."
            )
        citation["context_begin"] = matches[0]["begin_idx"]
        citation["context_end"] = matches[0]["end_idx"]

    return augmented_citation_info



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Common utility functions for internal use by the library and its tests.
"""

# Standard
import contextlib
import itertools
import json
import logging
import os
import re
import uuid

# Third Party
import pydantic

# First Party
from granite_common.base.types import (
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
)

NLTK_INSTALL_INSTRUCTIONS = """
Please install nltk with:
    pip install nltk
In some environments you may also need to manually download model weights with:
    python -m nltk.downloader punkt_tab
See https://www.nltk.org/install.html#installing-nltk-data for more detailed 
instructions."""


@contextlib.contextmanager
def import_optional(extra_name: str):
    """Context manager to handle optional imports"""
    try:
        yield
    except ImportError as err:
        logging.warning(
            "%s.\nHINT: You may need to pip install %s[%s]",
            err,
            __package__,
            extra_name,
        )
        raise


@contextlib.contextmanager
def nltk_check(feature_name: str):
    """Variation on import_optional for nltk.

    :param feature_name: Name of feature that requires NLTK"""
    try:
        yield
    except ImportError as err:
        raise ImportError(
            f"'nltk' package not installed. This package is required for "
            f"{feature_name} in the 'granite_io' library."
            f"{NLTK_INSTALL_INSTRUCTIONS}"
        ) from err


def find_substring_in_text(substring: str, text: str) -> list[int]:
    """
    Given two strings - substring and text - find and return all
    matches of substring within text. For each match return its begin and end index
    """
    span_matches = []

    matches_iter = re.finditer(re.escape(substring), text)
    for match in matches_iter:
        span_matches.append({"begin_idx": match.start(), "end_idx": match.end()})

    return span_matches


def random_uuid() -> str:
    """:returns: hexadecimal data suitable to use as a unique identifier"""
    return str(uuid.uuid4())


def load_transformers_lora(local_or_remote_path):
    """
    AutoModelForCausalLM.from_pretrained() is supposed to auto-load base models if you
    pass it a LoRA adapter's config, but that auto-loading is very broken as of 8/2025.
    Workaround powers activate!

    Only works if ``transformers`` and ``peft`` are installed

    :returns: Tuple of LoRA model and tokenizer
    """
    with import_optional("peft"):
        # Third Party
        import peft
        import transformers
    local_model_dir = local_or_remote_path
    if not os.path.exists(local_model_dir):
        raise NotImplementedError("TODO: Talk to hugging face hub")
    with open(f"{local_model_dir}/adapter_config.json", encoding="utf-8") as f:
        adapter_config = json.load(f)
    base_model_name = adapter_config["base_model_name_or_path"]
    base_model = transformers.AutoModelForCausalLM.from_pretrained(base_model_name)
    tokenizer = transformers.AutoTokenizer.from_pretrained(base_model_name)
    model = peft.PeftModel.from_pretrained(base_model, local_model_dir)
    return model, tokenizer


def chat_completion_request_to_transformers_inputs(
    request, tokenizer=None, model=None, constrained_decoding_prefix=None
) -> tuple[dict, dict, dict]:
    """
    Translate an OpenAI-style chat completion request into an input for a Transformers
    ``generate()`` call.

    :param request: Request as parsed JSON or equivalent dataclass
    :param tokenizer: Pointer to the HuggingFace tokenizer that will be used to handle
        this request. Only required if the request uses constrained decoding.
    :param model: Pointer to the HuggingFace model that will be used to handle
        this request. Only required if the request uses constrained decoding.
    :param constrained_decoding_prefix: Optional generation prefix to append to the
        prompt

    :returns: Tuple of:
        * kwargs to pass to generation
        * Additional stuff to pass to generate_with_transformers
    """
    with import_optional("torch"):
        # Third Party
        import torch

    if isinstance(request, pydantic.BaseModel):
        request = request.model_dump()

    generate_input = {
        # Always return dict, else downstream code will need lots type checks
        "return_dict_in_generate": True
    }

    tokenizer_input = {
        "conversation": request["messages"],
        "add_generation_prompt": True,
    }

    # pylint: disable=unsupported-membership-test
    if (
        request.get("extra_body") is not None
        and request["extra_body"].get("documents") is not None
    ):
        tokenizer_input["documents"] = request["extra_body"]["documents"]

    input_tokens = tokenizer.apply_chat_template(**tokenizer_input, return_tensors="pt")

    # generate() will fail with many different creative error messages if tokens aren't
    # on the right device.
    input_tokens = input_tokens.to(model.device)
    generate_input["input_tokens"] = input_tokens

    # The generate() method sometimes needs to know what is the integer ID
    # of the padding token, and for some reason this critical piece of information
    # isn't included in the serialized model. We get it from the tokenizer.
    # And of course some tokenizers don't set this parameter, in which case
    # we use the end of string token and hope for the best.
    pad_token_id = tokenizer.pad_token_id
    if pad_token_id is None:
        pad_token_id = tokenizer.eos_token_id
    if pad_token_id is None:
        # Raise an error here because the some branches of the generate
        # method won't complain about an invalid value of this parameter,
        # while others will raise a cryptic exception from deep within
        # their beam search code.
        raise ValueError(f"Couldn't figure out padding token for tokenizer {tokenizer}")
    generate_input["pad_token_id"] = pad_token_id

    # Make sure you specify this parameter explicitly, or you will have
    # a bad time.
    generate_input["eos_token_id"] = (tokenizer.eos_token_id,)

    other_input = {}

    if "logprobs" in request and request["logprobs"]:
        generate_input["output_scores"] = True

    if request.get("top_logprobs") is not None:
        # Transformers has no notion of top_logprobs. Pass it through so our own post-
        # processing code can deal with it on the other side.
        other_input["top_logprobs"] = request["top_logprobs"]

    if request.get("max_completion_tokens") is not None:
        generate_input["max_new_tokens"] = request["max_completion_tokens"]

    if (
        request.get("extra_body") is not None
        and request["extra_body"].get("guided_json") is not None
    ):
        # Constrained decoding in Hugging Face requires using a third-party library
        # to create a callback function to be invoked from inside generate()
        with import_optional("xgrammar"):
            # Third Party
            import xgrammar as xgr
        if tokenizer is None:
            raise ValueError(
                "Request specifies constrained decoding, but no "
                "tokenizer object was passed to this function."
            )
        if model is None:
            raise ValueError(
                "Request specifies constrained decoding, but no "
                "tokenizer object was passed to this function."
            )

        # Different parts of a Hugging Face model will have different opinions about
        # the number of tokens in the tokenizer's vocabulary, because of course they do.
        # Gather together all the possibilities and pick the biggest one.
        vocab_size = max(tokenizer.vocab_size, len(tokenizer), model.vocab_size)

        tokenizer_info = xgr.TokenizerInfo.from_huggingface(
            tokenizer,
            vocab_size=vocab_size,
        )
        grammar_compiler = xgr.GrammarCompiler(tokenizer_info)
        compiled_grammar = grammar_compiler.compile_json_schema(
            request["extra_body"]["guided_json"]
        )
        logits_processor = xgr.contrib.hf.LogitsProcessor(compiled_grammar)

        # The "logits_processor" argument to generate() must be a list.
        generate_input["logits_processor"] = [logits_processor]

        if constrained_decoding_prefix is not None:
            # Some models generate boilerplate before getting to the place where the
            # logits processor should activate. Append that boilerplate to the prompt,
            # since the logits processor we just created will
            addl_tokens = tokenizer(
                constrained_decoding_prefix, return_tensors="pt"
            ).to(model.device)["input_ids"]
            generate_input["input_tokens"] = torch.cat(
                [generate_input["input_tokens"], addl_tokens], dim=1
            )

    # Translate beam search parameters
    if request.get("temperature") is not None:
        if request["temperature"] == 0.0:
            # No beam search
            generate_input["do_sample"] = False
        else:
            # Beam search
            generate_input["do_sample"] = True
            generate_input["temperature"] = request["temperature"]

    if request.get("n") is not None:
        generate_input["num_return_sequences"] = request["n"]

    for param in (
        "top_k",
        "top_p",
    ):
        if request.get(param) is not None:
            generate_input[param] = request[param]

    return generate_input, other_input


def generate_with_transformers(
    tokenizer,
    model,
    generate_input: dict,
    other_input: dict,
) -> ChatCompletionResponse:
    """
    All the extra steps necessary to call the :func:`generate()` method of a
    Transformers model and get back usable results, rolled into a single function.

    There are quite a few extra steps.

    :param tokenizer: Tokenizer for the model, required at several stages of generation
    :param model: Initialized model object.
    :param generate_input: Parameters to pass to the generate() method, usually
        generated by :func:`chat_completion_request_to_transformers_inputs()`
    :param other_input: Additional kwargs that
        :func:`chat_completion_request_to_transformers_inputs()` added to encompass
        aspects of the original request that Transformers APIs don't handle natively.

    :returns: A chat completion response in OpenAI format
    """
    with import_optional("torch"):
        # Third Party
        import torch

    # Input tokens must be passed to generate() as a positional argument, not a named
    # argument.
    input_tokens = generate_input["input_tokens"]
    generate_input = generate_input.copy()
    del generate_input["input_tokens"]

    generate_result = model.generate(input_tokens, **generate_input)

    # Result is a a 2D tensor of shape (num responses, prompt + max generated tokens)
    # containing tokens, plus a tuple of <max generated tokens> tensors of shape
    # (num beams, vocab size) containing scores.
    # This is of course not a usable format for downstream processing.
    # Start by stripping off the prompt, leaving us with a tensor of shape
    # (num responses, max generated tokens)
    num_prompt_tokens = input_tokens.shape[1]
    num_responses = generate_result.sequences.shape[0]
    generated_tokens = generate_result.sequences[:, num_prompt_tokens:]

    generated_scores = (
        None
        if generate_result.scores is None
        else (torch.stack(generate_result.scores).swapaxes(0, 1)[:num_responses])
    )

    # Iterate over the responses, stripping off EOS tokens
    choices = []
    for i in range(num_responses):
        response_tokens = generated_tokens[i]

        if tokenizer.eos_token_id in response_tokens:
            # Strip off everything after the first EOS token.
            # Pytorch syntax for finding the first EOS is a bit funky.
            eos_ix = (
                (response_tokens == tokenizer.eos_token_id)
                .nonzero(as_tuple=True)[0]
                .item()
            )
            response_tokens = response_tokens[:eos_ix]

        response_string = tokenizer.decode(response_tokens)

        # The decode() method doesn't return offsets.
        # The only supported API to get offsets is to retokenize the string and hope you
        # get back the same tokenization.
        # This supported API doesn't work reliably, so we fall back on the unsupported
        # method of pulling token lengths out of the tokenizer.
        ends = list(
            itertools.accumulate(
                [len(s) for s in tokenizer.batch_decode(response_tokens)]
            )
        )
        begins = [0] + ends[:-1]
        token_offsets = list(zip(begins, ends, strict=True))

        if generated_scores is None:
            logprobs_content = None
        else:
            response_scores = generated_scores[i]

            # Scores come back as raw logits. You need to decode them to produce
            # logprobs. For consistency with the OpenAI output format, we need to
            # decode twice: Once to get the probability of the returned token and a
            # second time to get the top k logprobs. As with the OpenAI APIs, the
            # returned token may or may not be included in the top k results.
            all_logprobs = torch.log_softmax(response_scores.to(torch.float32), 1)
            chosen_token_logprobs = [
                all_logprobs[token_ix][response_tokens[token_ix]].item()
                for token_ix in range(len(response_tokens))
            ]
            token_strings = [response_string[begin:end] for begin, end in token_offsets]
            token_bytes = [list(s.encode("utf-8")) for s in token_strings]

            # Transformers has no notion of top-k logprobs, so the parameter that
            # triggers that post-processing is passed via other_input.
            if "top_logprobs" not in other_input:
                top_logprobs = [[] for _ in range(len(token_strings))]
            else:  # if "top_logprobs" in other_input:
                top_k_values, top_k_indices = torch.topk(
                    torch.nan_to_num(all_logprobs, float("-inf")),
                    other_input["top_logprobs"],
                )
                top_k_token_strs = [
                    [tokenizer.decode(t) for t in row_i] for row_i in top_k_indices
                ]
                top_logprobs = [
                    [
                        {
                            "token": s,
                            "bytes": list(s.encode("utf8")),
                            "logprob": lp.item(),
                        }
                        for s, lp in zip(strs, lps, strict=True)
                    ]
                    for strs, lps in zip(top_k_token_strs, top_k_values, strict=True)
                ]

            logprobs_content = [
                {
                    "token": token_strings[i],
                    "bytes": token_bytes[i],
                    "logprob": chosen_token_logprobs[i],
                    "top_logprobs": top_logprobs[i],
                }
                for i in range(len(response_tokens))
            ]

        response_choice_value = {
            "index": i,
            "message": {"content": response_string, "role": "assistant"},
        }
        if logprobs_content is not None:
            response_choice_value["logprobs"] = {"content": logprobs_content}
        response_choice = ChatCompletionResponseChoice.model_validate(
            response_choice_value
        )
        choices.append(response_choice)

    return ChatCompletionResponse(choices=choices)



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Input and output processing for the Granite 3.3 family of models.
"""

# Local
from .input import Granite33InputProcessor
from .output import Granite33OutputProcessor
from .types import Granite33ChatCompletion

__all__ = (
    "Granite33ChatCompletion",
    "Granite33InputProcessor",
    "Granite33OutputProcessor",
)



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Classes and functions that implement common aspects of output processing for all
LoRA adapters in IBM's `rag-agent-lib` library of intrinsics.
"""

# Standard
from typing import Any
import abc
import collections
import copy
import enum
import json
import math
import pathlib

# Third Party
import pydantic

# First Party
from granite_common.base.io import ChatCompletionResultProcessor
from granite_common.base.types import (
    ChatCompletion,
    ChatCompletionLogProb,
    ChatCompletionLogProbs,
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
    Document,
)

# Local
from . import json_util
from .input import sentence_delimiter
from .util import make_config_dict


class _MappingType(enum.Enum):
    PASSTHRU = 1
    """Pass through the value from the model output using the type of the raw output
    schema"""


class TransformationRule(abc.ABC):
    """Base class for transformation rules to apply to JSON outputs of intrinsics."""

    YAML_NAME = None
    """Subclasses should set this to the name of the rule in YAML config files."""

    def __init__(self, config: dict, input_path_expr: list[str | int | None]):
        """
        :param config: Configuration of the parent output processor, as parsed YAML.
        :param input_path_expr: Path expression that matches all instances of the field
            that this rule transforms. Elements can be strings for object fields,
            ints for list indices, or ``None`` for wildcard matches.
        """
        self.config = config
        self.input_path_expr = input_path_expr

    def _type_check(self, attr_name, attr_value, expected_type):
        """Convenience method for subclasses to type-check an argument from the YAML
        config file."""
        if not isinstance(attr_value, expected_type):
            raise TypeError(
                f"'{attr_name}' parameter for '{self.rule_name()}' rule "
                f" must be of type {expected_type}."
                f"list. Received value {attr_value}"
            )

    def _is_input_path(self, path: tuple) -> bool:
        """
        :param path: JSON path as returned by :func:`scalar_paths()`

        :returns: True if this rule should be applied to the indicated path.
        """
        if len(path) != len(self.input_path_expr):
            return False
        for expr_elem, path_elem in zip(self.input_path_expr, path, strict=True):
            # None means "wildcard"
            if expr_elem is not None and path_elem != expr_elem:
                return False
        return True

    def _matching_paths(self, parsed_json: Any) -> list[tuple]:
        """
        :param parsed_json: Output of running model results through
            :func:`json.loads()`, plus applying zero or more transformation rules.
        :returns: List of paths within ``parsed_json`` that match this rule's input
            path spec.
        """
        return [p for p in json_util.all_paths(parsed_json) if self._is_input_path(p)]

    def rule_name(self) -> str:
        if self.YAML_NAME is None:
            raise ValueError(f"Attempted to fetch missing rule name for {type(self)}")
        return self.YAML_NAME

    # pylint: disable=unused-argument
    def _prepare(
        self,
        parsed_json: Any,
        reparsed_json: Any,
        logprobs: ChatCompletionLogProbs | None,
        chat_completion: ChatCompletion | None,
    ) -> dict:
        """
        Subclasses may override this method to prepare data structures that should be
        computed once per output.

        :returns: Dict that will be passed to all calls to :func:`self._transform()`
        """
        return {}

    def apply(
        self,
        parsed_json: Any,
        reparsed_json: Any,
        logprobs: ChatCompletionLogProbs | None,
        chat_completion: ChatCompletion | None,
    ) -> Any:
        """
        Main entry point.

        :param parsed_json: Output of running model results through
            :func:`json.loads()`, plus applying zero or more transformation rules.
        :param reparsed_json: Output of running the same model results through
            :func:`json_util.reparse_json_with_offsets()`.
        :param logprobs: Optional logprobs result associated with the original model
            output string, or ``None`` of no logprobs were present.
        :param chat_completion: The chat completion request that produced this output.
        :returns: Transformed copy of ``parsed_json`` after applying this rule.
        """
        paths = self._matching_paths(parsed_json)
        prepare_output = self._prepare(
            parsed_json, reparsed_json, logprobs, chat_completion
        )

        # print(f"{self.rule_name()}: Matching paths are {paths}")

        # If we get here, we need to modify a JSON object or array
        # Don't modify input
        result = copy.deepcopy(parsed_json)
        for path in paths:
            result = self._apply_at_path(result, path, prepare_output)
        return result

    @abc.abstractmethod
    def _apply_at_path(self, result: Any, path: tuple, prepare_output: dict) -> Any:
        """
        Subclasses should modify this

        :param result: Parsed JSON representation of the transformed output at the
            current stage of transformation.  A copy of the original.
        :param path: Current path to transform locally
        :param prepare_output: Dictionary of global data that this object's
            :func:`self._prepare()` method has set aside

        :returns: A modified version of ``result``, which may be modified in place or
            a fresh copy.
        """
        raise NotImplementedError()


class InPlaceTransformation(TransformationRule):
    """
    Base class for ``TransformationRule``s that replace values in place in the source
    JSON. The values replaced can be a scalar, object, or list.
    """

    def _apply_at_path(self, result: Any, path: tuple, prepare_output: dict) -> Any:
        original_value = json_util.fetch_path(result, path)
        transformed_value = self._transform(original_value, path, prepare_output)
        result = json_util.replace_path(result, path, transformed_value)
        return result

    @abc.abstractmethod
    def _transform(self, value: Any, path: tuple, prepare_output: dict) -> Any:
        """
        Subclasses should override this method to transform a single scalar value
        from a larger JSON expression.

        :param value: Original value pulled out of the JSON expression, with position
            information attached to any embedded scalars.
        :param path: Location in the input JSON where the value was found
        :param prepare_output: Results from calling :func:`self._prepare()`

        :returns: New value for the indicated element of the JSON expression.
        """
        raise NotImplementedError()


class AddFieldsTransformation(TransformationRule):
    """
    Base class for ``TransformationRule``s that add one or more values adjacent to
    an existing value in the source JSON.
    """

    def _apply_at_path(self, result: Any, path: tuple, prepare_output: dict) -> Any:
        """
        Subclasses should modify this

        :param result: Parsed JSON representation of the transformed output at the
            current stage of transformation.  A copy of the original.
        :param path: Current path to transform locally
        :param prepare_output: Dictionary of global data that this object's
            :func:`self._prepare()` method has set aside

        :returns: A modified version of ``result``, which may be modified in place or
            a fresh copy.
        """
        if len(path) == 0:
            raise ValueError(
                "Expected path to field of JSON object, but received zero-length path."
            )

        parent_path = path[:-1]
        parent_object = json_util.fetch_path(result, parent_path)
        if not isinstance(parent_object, dict):
            raise TypeError(
                f"Expected JSON object at path {parent_path} but found value of type "
                f"{type(parent_object)}"
            )

        original_value = parent_object[path[-1]]
        new_values = self._transform(original_value, path, prepare_output)

        # Make a copy, just in case.
        new_parent = parent_object.copy() | new_values
        result = json_util.replace_path(result, parent_path, new_parent)
        return result

    @abc.abstractmethod
    def _transform(self, value: Any, path: tuple, prepare_output: dict) -> dict:
        """
        Subclasses should override this method to transform a single scalar value
        from a larger JSON expression.

        :param value: Original value pulled out of the JSON expression, with position
            information attached to any embedded scalars.
        :param path: Location in the input JSON where the value was found
        :param prepare_output: Results from calling :func:`self._prepare()`

        :returns: Mapping from name of added field to value.
        """
        raise NotImplementedError()


##################################################
# Rule implementation classes start here


class TokenToFloat(InPlaceTransformation):
    """
    Transformation rule that decodes token logprobs to a floating point number.

    The floating point number replaces the original categorical value in the JSON.
    """

    YAML_NAME = "likelihood"

    def __init__(
        self,
        config: dict,
        input_path_expr: list[str | int | None],
        /,
        categories_to_values: dict[str | int | bool, float] | None = None,
    ):
        """
        :param categories_to_values: Mapping from categorical labels to floating-point
            values.
        :type categories_to_values: dict[str | int | bool, float]
        """
        super().__init__(config, input_path_expr)
        self.categories_to_values = categories_to_values

    def _prepare(
        self,
        parsed_json: Any,
        reparsed_json: Any,
        logprobs: ChatCompletionLogProbs | None,
        chat_completion: ChatCompletion | None,
    ) -> dict:
        if logprobs is not None and not isinstance(logprobs, ChatCompletionLogProbs):
            raise TypeError(
                f"Expected ChatCompletionLogProbs, but received {type(logprobs)}"
            )
        if logprobs is None:
            raise TypeError("This rule requires logprobs.  Received None for logprobs.")
        begin_to_token = json_util.make_begin_to_token_table(logprobs)

        return {
            "begin_to_token": begin_to_token,
            "reparsed_json": reparsed_json,
            "logprobs": logprobs,
        }

    def _transform(self, value: Any, path: tuple, prepare_output: dict) -> Any:
        # Retrieve values that are computed during self._prepare()
        begin_to_token = prepare_output["begin_to_token"]
        logprobs = prepare_output["logprobs"]
        reparsed_json = prepare_output["reparsed_json"]

        json_literal = json_util.fetch_path(reparsed_json, path)
        if not isinstance(json_literal, json_util.JsonLiteralWithPosition):
            raise TypeError(
                f"Expected literal with position, but received '{value}' "
                f"of type {type(value)}"
            )
        if value != json_literal.value:
            # Sanity check: Can't apply this rule on a path for which the tokens for
            # the original logprobs are no longer present.
            raise ValueError(
                f"At path {path}, reparsed value '{json_literal}' differs from "
                f"current value '{value}'. This rule cannot decode logprobs under this "
                f"circumstance."
            )

        value_str_offset = json_literal.begin
        if isinstance(json_literal.value, str):
            # Skip double quote at beginning of string literal
            value_str_offset += 1

        if value_str_offset not in begin_to_token:
            # Categorical value didn't start on a token boundary. Check whether the
            # value was the suffix of a token.
            prev_token_offset = value_str_offset
            while prev_token_offset > 0 and prev_token_offset not in begin_to_token:
                prev_token_offset -= 1
            prev_token_ix = begin_to_token[prev_token_offset]
            prev_token = logprobs.content[prev_token_ix].token
            if prev_token.endswith(json.dumps(json_literal.value)):
                # Can't decode other tokens since we don't have logprobs for the
                # beginning of the literal.
                return self.categories_to_values[json_literal.value]

            raise ValueError(
                f"Value '{json_literal}' starts at position "
                f"{value_str_offset}, "
                f"but there is no token at that position."
            )

        first_token_ix = begin_to_token[value_str_offset]
        values = []
        weights = []

        # Decode top token.
        # Assume that probability of first token == probability of entire literal
        if json_literal.value in self.categories_to_values:
            values.append(self.categories_to_values[json_literal.value])
            weights.append(math.exp(logprobs.content[first_token_ix].logprob))

        # Decode remaining tokens.
        # Here we assume that the first category that shares a prefix with the token is
        # what the completion would have been had that token been the top-1.
        top_logprob: ChatCompletionLogProb
        for top_logprob in logprobs.content[first_token_ix].top_logprobs:
            if top_logprob.token == logprobs.content[first_token_ix].token:
                # Some inference engines will output the top-1 token both in logprobs
                # and in top_logprobs some of the time. Don't double-count when that
                # happens.
                continue
            for category, value_for_category in self.categories_to_values.items():
                if str(category).startswith(top_logprob.token):
                    # Use the first prefix match
                    values.append(value_for_category)
                    weights.append(math.exp(top_logprob.logprob))
                    break

        # Make the weights sum to 1 and return weighted sum, aka expected value
        if len(values) == 0:
            # No match --> default to 0
            return 0.0
        total_weight = sum(weights)
        weights = [w / total_weight for w in weights]
        return sum(w * v for w, v in zip(weights, values, strict=True))


def _desplit_sentences(
    target_text: str,
    tag: str,
    first_sentence_num: int,
) -> dict[str, list]:
    """Subroutine of :class:`DecodeSentences` rule. Undoes the sentence splitting
    that we sometimes do during input processing.

    :param target_text: Text that has been split into sentences by inserting sentence
        boundary markers.
    :param tag: String such as that appears in every sentence boundary marker, e.g.
        "i" => "<i123>"
    :param first_sentence_num: Number we expect to see in the first sentence boundary
        marker in ``target_text``.

    :returns: Self-describing dictionary of lists.
    """
    begins = []
    ends = []
    texts = []

    sentence_num = first_sentence_num

    # There is always at least one sentence if the text is encoded correctly.
    delimiter = sentence_delimiter(tag, sentence_num)
    delimiter_loc = target_text.find(delimiter, 0)
    if delimiter_loc == -1:
        raise ValueError(
            f"First sentence delimiter '{delimiter}' not found in "
            f"target string '{target_text}'"
        )
    begin = 0
    tagged_begin = delimiter_loc + len(delimiter)

    while True:
        # Loop invariant: We are looking for the end of sentence <sentence_num>.
        #   <begin> is positioned at the beginning of <sentence_num>,
        #   immediately after the delimiter for that sentence.

        # Delimiter string occurs at the BEGINNING of every sentence.
        # Check for delimiter of next sentence.
        delimiter = sentence_delimiter(tag, sentence_num + 1)
        delimiter_loc = target_text.find(delimiter, tagged_begin)
        if delimiter_loc == -1:
            # No more sentence markers
            begins.append(begin)
            # Begin + (remaining characters in tagged text)
            ends.append(begin + (len(target_text) - tagged_begin))
            texts.append(target_text[tagged_begin:])
            break
        begins.append(begin)
        begin += delimiter_loc
        ends.append(begin)
        new_tagged_begin = delimiter_loc + len(delimiter)
        texts.append(target_text[tagged_begin:delimiter_loc])
        tagged_begin = new_tagged_begin
        sentence_num += 1

    return {"begins": begins, "ends": ends, "texts": texts}


class DecodeSentences(AddFieldsTransformation):
    """
    Transformation rule that decodes references to sentences by number into begin, end,
    text tuples.
    """

    YAML_NAME = "decode_sentences"

    def __init__(
        self,
        config: dict,
        input_path_expr: list[str | int | None],
        /,
        source: str,
        output_names: dict,
    ):
        """
        :param source: Name of the location to look for sentences; can be "last_message"
             or "documents".
        :param output_names: Names of new result fields to add
        """
        super().__init__(config, input_path_expr)

        allowed_sources = ("last_message", "documents")
        if source not in allowed_sources:
            raise ValueError(
                f"'source' argument must be one of {allowed_sources}. "
                f"Received '{source}'"
            )
        self.source = source

        if not isinstance(output_names, dict):
            raise TypeError(
                f"Expected mapping for output_names, but received {output_names}"
            )
        for k in output_names:
            if source == "documents" and k == "document_id":
                continue
            if k not in ("begin", "end", "text"):
                raise ValueError(f"Unexpected key '{k}' in output_names")

        # Each of these attributes is set to None if not present in the YAML file
        self.begin_name = output_names.get("begin")
        self.end_name = output_names.get("end")
        self.text_name = output_names.get("text")
        self.document_id_name = output_names.get("document_id")

        if config["docs_as_message"] and config["docs_as_message"] != "json":
            raise ValueError(
                f"Decoding sentences from message with document "
                f"encoding method '{config['docs_as_message']}' is "
                f"not yet supported."
            )

    def _prepare(
        self,
        parsed_json: Any,
        reparsed_json: Any,
        logprobs: ChatCompletionLogProbs | None,
        chat_completion: ChatCompletion | None,
    ) -> dict:
        if chat_completion is None:
            raise ValueError(
                f"No chat completion request object provided. "
                f"'{self.rule_name()}' rule requires this object."
            )

        if self.source == "documents":
            tag = self.config["sentence_boundaries"]["documents"]
            if tag is None:
                raise ValueError(
                    f"'{self.rule_name()}' attempting to decode document sentences, "
                    f"but 'sentence_boundaries' section of config file is missing "
                    f"the entry that tells how to tag document sentence boundaries."
                )

            if not self.config["docs_as_message"]:
                # Most common path: Documents from extra_body
                documents = chat_completion.extra_body.documents
            else:
                # Model requires documents in a user message. Decode the message.
                # Currently onli JSON format is supported.
                if self.config["docs_as_message"] != "json":
                    raise ValueError(
                        f"Unsupported doc type {self.config['docs_as_message']}"
                    )
                documents_json = json.loads(chat_completion.messages[0].content)
                documents = [Document.model_validate(d) for d in documents_json]

            if documents is None:
                documents = []

            # De-split the sentences in each document in turn. Sentence numbers
            # start at zero on the first document and continue in subsequent documents.
            begins = []
            ends = []
            texts = []
            document_ids = []

            next_sentence_num = 0
            for d in documents:
                local_results = _desplit_sentences(d.text, tag, next_sentence_num)
                num_local_sentences = len(local_results["begins"])
                begins.extend(local_results["begins"])
                ends.extend(local_results["ends"])
                texts.extend(local_results["texts"])
                document_ids.extend([d.doc_id] * num_local_sentences)
                next_sentence_num += num_local_sentences

            return {
                "begins": begins,
                "ends": ends,
                "texts": texts,
                "document_ids": document_ids,
            }
        if self.source == "last_message":
            tag = self.config["sentence_boundaries"]["last_message"]
            if tag is None:
                raise ValueError(
                    f"'{self.rule_name()}' attempting to decode the last message, "
                    f"but 'sentence_boundaries' section of config file is missing "
                    f"the entry that tells how to tag message sentence boundaries."
                )

            # Use second-to-last turn if the input processing added an instruction turn
            message_ix = -2 if self.config["instruction"] else -1
            target_text = chat_completion.messages[message_ix].content

            return _desplit_sentences(target_text, tag, 0)

        raise ValueError(f"Unexpected source string '{self.source}'")

    def _transform(self, value: Any, path: tuple, prepare_output: dict) -> dict:
        # Unpack global values we set aside during the prepare phase
        begins = prepare_output["begins"]
        ends = prepare_output["ends"]
        texts = prepare_output["texts"]
        document_ids = prepare_output.get("document_ids")

        if not isinstance(value, int):
            raise TypeError(
                f"Expected integer sentence number at path {path}, but "
                f"found non-integer value {value} of type {type(value)}"
            )
        sentence_num = value

        result = {}
        if self.begin_name is not None:
            result[self.begin_name] = begins[sentence_num]
        if self.end_name is not None:
            result[self.end_name] = ends[sentence_num]
        if self.text_name is not None:
            result[self.text_name] = texts[sentence_num]
        if self.document_id_name is not None:
            result[self.document_id_name] = document_ids[sentence_num]
        return result


class Explode(InPlaceTransformation):
    """Turn each row in a list of records into zero or more rows by expanding
    the elements of a list-valued attribute."""

    YAML_NAME = "explode"

    def __init__(self, config, input_path_expr, /, target_field):
        """
        :param config: Parsed YAML config for IO processing
        :param input_path_expr: Path expression for the list of records to explode
        :param target: Name of list-valued field within each record.
        """
        super().__init__(config, input_path_expr)
        self.target_field = target_field

    def _transform(self, value, path, prepare_output):
        # Lots of error checking here because the input has no type constraints.
        if not isinstance(value, list):
            raise TypeError(
                f"Matching element at path {path} is not a list. "
                f"Matching paths of an '{self.rule_name()}' rule must be "
                f"lists."
            )
        result = []

        for i, element in enumerate(value):
            if not isinstance(element, dict):
                raise TypeError(
                    f"Element {i} of list at path {path} is not a record. "
                    f"'{self.rule_name()}' rule requires a list of records. "
                    f"Element is: {element}"
                )
            if self.target_field not in element:
                raise ValueError(
                    f"Element {i} of list at path {path} does not contain target field "
                    f"'{self.target_field}' for '{self.rule_name()}' rule. "
                    f"Element is: {element}"
                )
            to_explode = element[self.target_field]
            if not isinstance(to_explode, list):
                raise ValueError(
                    f"Target field '{self.target_field}' of element {i} of list at "
                    f"path {path} does not contain a list. "
                    f"'{self.rule_name()}' rule requires a list-valued attribute. "
                    f"Element is: {element}"
                )

            # Now that we've verified all the invariants, we can get down to the
            # business of constructing output records.
            for target_value in to_explode:
                # Use a single dictionary comprehension to ensure that field order
                # does not change.
                result.append(
                    {
                        k: target_value if k == self.target_field else v
                        for k, v in element.items()
                    }
                )
        return result


class DropDuplicates(InPlaceTransformation):
    """Remove duplicate records from a list of records."""

    YAML_NAME = "drop_duplicates"

    def __init__(self, config, input_path_expr, /, target_fields):
        """
        :param config: Parsed YAML config for IO processing
        :param input_path_expr: Path expression for the list of records to explode
        :param target_fields: Names of fields to use for determining whether two records
            are duplicates.
        """
        super().__init__(config, input_path_expr)
        self.target_fields = target_fields

    def _transform(self, value, path, prepare_output):
        # Lots of error checking here because the input has no type constraints.
        if not isinstance(value, list):
            raise TypeError(
                f"Matching element at path {path} is not a list. "
                f"Matching paths of an '{self.rule_name()}' rule must be "
                f"lists."
            )

        # Build up a mapping from value of dedup key to last record with that key
        key_to_record = {}
        for i, element in enumerate(value):
            if not isinstance(element, dict):
                raise TypeError(
                    f"Element {i} of list at path {path} is not a record. "
                    f"'{self.rule_name()}' rule requires a list of records. "
                    f"Element is: {element}"
                )
            for t in self.target_fields:
                if t not in element:
                    raise TypeError(
                        f"Element {i} of list at path {path} is missing field '{t}'. "
                        f"'{self.rule_name()}' rule requires that the fields used for "
                        f"deduplication be present in all rows. "
                        f"Element is: {element}"
                    )
            key = (element[t] for t in self.target_fields)
            key_to_record[key] = element
        return list(key_to_record.values())


class Project(InPlaceTransformation):
    """Project records down to a specified set of fields.  Can also rename the
    retained fields."""

    YAML_NAME = "project"

    def __init__(self, config, input_path_expr, /, retained_fields):
        """
        :param config: Parsed YAML config for IO processing
        :param input_path_expr: Path expression for the list of records to explode
        :param retained_fields: Names of fields that remain after the projection. Can
            be either a list of fields or a mapping from retained field to new name of
            retained field.
        """
        super().__init__(config, input_path_expr)

        if not isinstance(retained_fields, list | dict):
            raise TypeError(
                f"'retained_fields' argument of '{self.rule_name()}' rule must be "
                f"either a list of fields or a mapping from retained field to new "
                f"name of retained field."
            )
        if isinstance(retained_fields, list):
            retained_fields = {f: f for f in retained_fields}
        self.retained_fields = retained_fields

    def _transform(self, value, path, prepare_output):
        # Lots of error checking here because the input has no type constraints.
        if isinstance(value, dict):
            # Single record
            return {self.retained_fields[k]: value.get(k) for k in self.retained_fields}
        if isinstance(value, list):
            return [
                {self.retained_fields[k]: element.get(k) for k in self.retained_fields}
                for element in value
            ]
        raise TypeError(
            f"Matching element at path {path} is not a list or a dictionary. "
            f"Matching paths of an '{self.rule_name()}' rule must contain "
            f"one or more records."
        )


class Nest(InPlaceTransformation):
    """Convert a value within a JSON structure into a record with a single field."""

    YAML_NAME = "nest"

    def __init__(self, config, input_path_expr, /, field_name):
        """
        :param config: Parsed YAML config for IO processing
        :param input_path_expr: Path expression for the values to nest
        :param field_name: name of the single field in the JSON object that this rule
            will wrap around each matching value.
        """
        super().__init__(config, input_path_expr)
        self._type_check("field_name", field_name, str)
        self.field_name = field_name

    def _transform(self, value, path, prepare_output):
        # Note that we don't check the type of the target value. This rule will happily
        # nest a collection of very different values. We may want to revisit this
        # design in the future. Or we might not.
        return {self.field_name: value}


class MergeSpans(InPlaceTransformation):
    """Merge adjacent spans into larger spans."""

    YAML_NAME = "merge_spans"

    def __init__(
        self,
        config,
        input_path_expr,
        /,
        group_fields: list,
        begin_field: str,
        end_field: str,
        text_field: str | None = None,
    ):
        """
        :param config: Parsed YAML config for IO processing
        :param input_path_expr: Path expression for the list of records to explode
        :param group_fields: List of fields that are used for grouping prior to merge
            spans.
        :param begin_field: Name of field that holds the begin offset of spans
        :param end_field: Name of field that holds the end offset of spans
        :param text_field: Optional field containing covered text strings that should
            be concatenated when spans are merged.
        """
        super().__init__(config, input_path_expr)
        self._type_check("group_fields", group_fields, list)
        self._type_check("begin_field", begin_field, str)
        self._type_check("end_field", end_field, str)
        if text_field is not None:
            self._type_check("text_field", text_field, str)

        self.group_fields = group_fields
        self.begin_field = begin_field
        self.end_field = end_field
        self.text_field = text_field
        self.expected_fields = set(
            self.group_fields + [self.begin_field, self.end_field]
        )
        if self.text_field is not None:
            self.expected_fields.add(self.text_field)

    def _transform(self, value, path, prepare_output):
        # Lots of error checking here because the input has no type constraints.
        if not isinstance(value, list):
            raise TypeError(
                f"Matching element at path {path} is not a list. "
                f"Matching paths of an '{self.rule_name()}' rule must be "
                f"lists."
            )

        if len(value) == 0:
            # Save the code below from having to deal with this corner case
            return value

        # Extract keys for sorting and index the original records by sort key
        sort_key_to_records = collections.defaultdict(list)
        sort_keys = set()
        for i, record in enumerate(value):
            # Lots of checks here because input does not have a known schema
            if not isinstance(record, dict):
                raise TypeError(
                    f"Element {i} of list at path {path} is not a record. "
                    f"'{self.rule_name()}' rule requires a list of records. "
                    f"Element is: {record}"
                )
            if record.get(self.begin_field) is None:
                raise ValueError(
                    f"Record at position {i} of list at path {path} does not contain "
                    f"a begin offset in the '{self.begin_field}' field. "
                    f"'{self.rule_name()}' rule does not handle null spans. "
                    f"Record is: {record}"
                )
            if record.get(self.end_field) is None:
                raise ValueError(
                    f"Record at position {i} of list at path {path} does not contain "
                    f"an end offset in the '{self.end_field}' field. "
                    f"'{self.rule_name()}' rule does not handle null spans. "
                    f"Record is: {record}"
                )
            if self.text_field is not None and record.get(self.text_field) is None:
                raise ValueError(
                    f"Record at position {i} of list at path {path} does not contain "
                    f"span text field '{self.text_field}' field. "
                    f"'{self.rule_name()}' rule does not handle null spans. "
                    f"Record is: {record}"
                )
            for field in record:
                if field not in self.expected_fields:
                    raise ValueError(
                        f"Record at position {i} of list at path {path} contains "
                        f"unexpected field '{field}'. "
                        f"'{self.rule_name()}' rule does not support fields other than "
                        f"the grouping fields and the fields of the span. "
                        f"Record is: {record}"
                    )
            sort_key = tuple(
                record.get(f) for f in self.group_fields + [self.begin_field]
            )
            sort_key_to_records[sort_key].append(record)
            sort_keys.add(sort_key)

        # Run through the sorted keys, merging spans
        result = []
        sorted_keys = sorted(sort_keys)
        key = sorted_keys[0]
        current_group = key[:-1]  # Last element is begin
        sorted_keys = sorted_keys[1:]
        records = sort_key_to_records[key]
        if len(records) != 1:
            raise ValueError(
                f"{len(records)} spans start at position "
                f"{records[0][self.begin_field]} of group {current_group}: {records}. "
                f"'{self.rule_name()}' rule only supports non-overlapping spans."
            )
        current_record = records[0].copy()  # Let's not modify the input, shall we?
        num_spans_merged = 1
        for key in sorted_keys:
            records = sort_key_to_records[key]
            if len(records) != 1:
                raise ValueError(
                    f"{len(records)} spans start at position "
                    f"{records[0][self.begin_field]} of group {current_group}: "
                    f"{records}. "
                    f"'{self.rule_name()}' rule only supports non-overlapping spans."
                )
            next_record = records[0]
            next_record_group = key[:-1]  # Last element is begin

            if (
                next_record_group != current_group
                or current_record[self.end_field] != next_record[self.begin_field]
            ):
                # Non-contiguous span.
                result.append(current_record)
                current_record = next_record.copy()
                current_group = next_record_group
                num_spans_merged = 1
            else:
                # Contiguous span. Extend current record.
                # Code above will have verified that only expected fields are present
                # and that begin, end, and text (if used) are always present.
                current_record[self.end_field] = next_record[self.end_field]
                if self.text_field is not None:
                    current_record[self.text_field] += next_record[self.text_field]
                num_spans_merged += 1

        # Don't forget the last output!
        result.append(current_record)
        return result


ALL_RULES = [
    # Try to keep these in alphabetical order, please
    DecodeSentences,
    DropDuplicates,
    Explode,
    MergeSpans,
    Nest,
    Project,
    TokenToFloat,
]
NAME_TO_RULE = {cls.YAML_NAME: cls for cls in ALL_RULES}

# END of rule implementations
############################################


class IntrinsicsResultProcessor(ChatCompletionResultProcessor):
    """General-purpose chat completion result processor for use with the models in the
    RAG Agent Library. Reads parameters of the model's input and output formats
    from a YAML configuration file and edits the input chat completion appropriately.
    """

    config: dict
    """Parsed YAML configuration file for the target intrinsic."""

    rules: list[TransformationRule]
    """Transformation rules that this object applies, in the order they are applied."""

    def __init__(
        self,
        /,
        config_file: str | pathlib.Path | None = None,
        config_dict: dict | None = None,
    ):
        """
        :param config_file: (optional) Location of YAML configuration file
        :param config_dict: (optional) Parsed contents of YAML configuration file
        """
        self.config = make_config_dict(config_file, config_dict)

        # Set up transformation rules for the target model's JSON output
        self.rules = []
        if self.config["transformations"]:
            for transform_spec in self.config["transformations"]:
                if transform_spec["type"] not in NAME_TO_RULE:
                    raise ValueError(
                        f"Unknown transformation rule '{transform_spec['type']}'. "
                        f"Available rules are: {list(NAME_TO_RULE)}"
                    )
                rule_cls = NAME_TO_RULE[transform_spec["type"]]
                input_path = transform_spec["input_path"]
                rule_kwargs = {
                    k: v
                    for k, v in transform_spec.items()
                    if k not in ("type", "input_path")
                }
                self.rules.append(rule_cls(self.config, input_path, **rule_kwargs))

    # pylint: disable=unused-argument
    def _transform_impl(
        self,
        chat_completion_response: ChatCompletionResponse | dict | pydantic.BaseModel,
        chat_completion: ChatCompletion | None = None,
    ) -> ChatCompletionResponse:
        transformed_choices = [
            self._transform_choice(c, chat_completion)
            for c in chat_completion_response.choices
        ]
        return chat_completion_response.model_copy(
            update={"choices": transformed_choices}
        )

    def _transform_choice(
        self,
        choice: ChatCompletionResponseChoice,
        chat_completion: ChatCompletion | None,
    ) -> ChatCompletionResponseChoice:
        # Parse JSON output twice: Once to verify valid JSON and once to compute offsets
        # Note that we don't currently check schema, as that would require an additional
        # library dependency.
        parsed_json = json.loads(choice.message.content)
        reparsed_json = json_util.reparse_json_with_offsets(choice.message.content)
        for rule in self.rules:
            parsed_json = rule.apply(
                parsed_json, reparsed_json, choice.logprobs, chat_completion
            )
        updated_message = choice.message.model_copy(
            update={"content": json.dumps(parsed_json)}
        )

        result = choice.model_copy(update={"message": updated_message})

        # Drop logprobs, since they should only be used by this function, and the tokens
        # referenced will no longer match the processed JSON value anyhow.
        # We may need to make this dropping configurable in the future.
        # Ok to modify in place because updated_message is a deep copy.
        if result.logprobs is not None:
            # Don't set the logprobs to None, unset it. There is a distinction in
            # Pydantic between these two states.
            del result.logprobs

        return result



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Common utility functions for this package.
"""

# Standard
import copy
import json
import os
import pathlib

# Third Party
import yaml

# Local
from .constants import (
    BASE_MODEL_TO_CANONICAL_NAME,
    RAG_INTRINSICS_LIB_REPO_NAME,
    YAML_JSON_FIELDS,
    YAML_OPTIONAL_FIELDS,
    YAML_REQUIRED_FIELDS,
)


def make_config_dict(
    config_file: str | pathlib.Path | None = None, config_dict: dict | None = None
) -> dict | None:
    """Common initialization code for reading YAML config files in factory classes.

    Also parses JSON fields.
    """
    if (config_file is None and config_dict is None) or (
        config_file is not None and config_dict is not None
    ):
        raise ValueError("Exactly one of config_file and config_dict must be set.")

    all_fields = sorted(YAML_REQUIRED_FIELDS + YAML_OPTIONAL_FIELDS)

    if config_dict:
        # Don't modify input
        config_dict = copy.deepcopy(config_dict)
    if config_file:
        with open(config_file, encoding="utf8") as file:
            config_dict = yaml.safe_load(file)

    # Validate top-level field names. No schema checking for YAML, so we need to do this
    # manually.
    for field in YAML_REQUIRED_FIELDS:
        if field not in config_dict:
            raise ValueError(f"Configuration is missing required field '{field}'")
    for name in config_dict:
        if name not in all_fields:
            raise ValueError(
                f"Configuration contains unexpected top-level field "
                f"'{name}'. Known top level fields are: {all_fields}"
            )
    for name in YAML_OPTIONAL_FIELDS:
        # Optional fields should be None if not present, to simplify downstream code.
        if name not in config_dict:
            config_dict[name] = None

    # Parse fields that contain JSON data.
    for name in YAML_JSON_FIELDS:
        if config_dict[name]:
            value = config_dict[name]
            # Users seem to be intent on passing YAML data through this function
            # multiple times, so we assume that values other than a string have already
            # been parsed by a previous call of this function.
            if isinstance(value, str):
                try:
                    config_dict[name] = json.loads(value)
                except json.JSONDecodeError as e:
                    raise ValueError(
                        f"Error parsing JSON in '{name}' field. Raw value was '{value}'"
                    ) from e

    return config_dict


def obtain_lora(
    intrinsic_name: str,
    target_model_name: str,
    /,
    alora: bool = False,
    repo_id=RAG_INTRINSICS_LIB_REPO_NAME,
    cache_dir: str | None = None,
    file_glob: str = "*",
) -> pathlib.Path:
    """
    Downloads a cached copy of a LoRA or aLoRA adapter from the
    [Granite Intrinsics Library](
    https://huggingface.co/ibm-granite/granite-3.3-8b-rag-agent-lib) if one is not
    already in the local cache.

    :param intrinsic_name: Short name of the intrinsic model, such as "certainty".
    :param target_model_name: Name of the base model for the LoRA or aLoRA adapter.
    :param alora: If ``True``, load aLoRA version of intrinsic; otherwise use LoRA
    :param repo_id: Optional name of Hugging Face Hub repository containing a collection
        of LoRA and/or aLoRA adapters for intrinsics.
        Default is to use rag-intrinsics-lib.
    :param cache_dir: Local directory to use as a cache (in Hugging Face Hub format),
        or ``None`` to use the Hugging Face Hub default location.
    :param file_glob: Only files that match this glob will be downloaded to the cache.

    :returns: the full path to the local copy of the specified (a)LoRA adapter.
    This path is suitable for passing to commands that will serve the adapter.
    """
    # Third Party
    import huggingface_hub

    # Normalize target model name.
    if target_model_name not in BASE_MODEL_TO_CANONICAL_NAME:
        raise ValueError(
            f"Unknown target model {target_model_name}. Known names are: "
            f"{list(BASE_MODEL_TO_CANONICAL_NAME.keys())}"
        )
    target_model_name = BASE_MODEL_TO_CANONICAL_NAME[target_model_name]

    lora_str = "alora" if alora else "lora"

    lora_subdir_name = f"{intrinsic_name}/{lora_str}/{target_model_name}"

    # Download just the files for this LoRA if not already present
    local_root_path = huggingface_hub.snapshot_download(
        repo_id=repo_id,
        allow_patterns=f"{lora_subdir_name}/{file_glob}",
        cache_dir=cache_dir,
    )
    lora_dir = pathlib.Path(local_root_path) / lora_subdir_name

    # Hugging Face Hub API will happily download nothing. Check whether that happened.
    if not os.path.exists(lora_dir):
        raise ValueError(
            f"Intrinsic '{intrinsic_name}' as "
            f"{'aLoRA' if alora else 'LoRA'} adapter on base model "
            f"'{target_model_name}' not found in "
            f"{repo_id} repository on Hugging Face Hub."
        )

    return lora_dir


def obtain_io_yaml(
    intrinsic_name: str,
    target_model_name: str,
    /,
    alora: bool = False,
    repo_id=RAG_INTRINSICS_LIB_REPO_NAME,
    cache_dir: str | None = None,
) -> pathlib.Path:
    """
    Downloads a cached copy of an ``io.yaml`` configuration file for an intrinsic in
    the [Granite Intrinsics Library](
    https://huggingface.co/ibm-granite/granite-3.3-8b-rag-agent-lib) if one is not
    already in the local cache.

    :param intrinsic_name: Short name of the intrinsic model, such as "certainty".
    :param target_model_name: Name of the base model for the LoRA or aLoRA adapter.
    :param alora: If ``True``, load aLoRA version of intrinsic; otherwise use LoRA
    :param repo_id: Optional name of Hugging Face Hub repository containing a collection
        of LoRA and/or aLoRA adapters for intrinsics.
        Default is to use rag-intrinsics-lib.
    :param cache_dir: Local directory to use as a cache (in Hugging Face Hub format),
        or ``None`` to use the Hugging Face Hub default location.

    :returns: the full path to the local copy of the specified (a)LoRA adapter.
    This path is suitable for passing to commands that will serve the adapter.
    """
    lora_dir = obtain_lora(
        intrinsic_name,
        target_model_name,
        alora,
        repo_id=repo_id,
        cache_dir=cache_dir,
        file_glob="io.yaml",
    )
    return lora_dir / "io.yaml"



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Constants used in code that is specific to the Granite 3.3 family of models.
"""

# Complete set of special tokens from the Granite 3.3 tokenizer
ALL_SPECIAL_TOKENS = [
    "<|end_of_text|>",
    "<fim_prefix>",
    "<fim_middle>",
    "<fim_suffix>",
    "<fim_pad>",
    "<filename>",
    "<gh_stars>",
    "<issue_start>",
    "<issue_comment>",
    "<issue_closed>",
    "<jupyter_start>",
    "<jupyter_text>",
    "<jupyter_code>",
    "<jupyter_output>",
    "<empty_output>",
    "<commit_before>",
    "<commit_msg>",
    "<commit_after>",
    "<reponame>",
    "<|start_of_role|>",
    "<|end_of_role|>",
    "<|tool_call|>",
    "<|start_of_cite|>",
    "<|end_of_cite|>",
    "<|start_of_plugin|>",
    "<|end_of_plugin|>",
]

# Delimiters for chain of thought output of Granite 3.3
COT_START = "<think>"
COT_END = "</think>"
RESPONSE_START = "<response>"
RESPONSE_END = "</response>"


# Delimiters for hallucination and citation output of Granite 3.3
CITATIONS_START = '{"id": "citation"}'
HALLUCINATIONS_START = '{"id": "hallucination"}'
CITE_START = "<|start_of_cite|>"
CITE_END = "<|end_of_cite|>"


# List of variations of the chain of thought start/end strings that the model is known
# to produce.
COT_START_ALTERNATIVES = [
    COT_START,
]
COT_END_ALTERNATIVES = [
    COT_END,
]

# String that a Granite 3.3 model must receive immediately after _SYSTEM_MESSAGE_START
# if there are both tools and RAG documents in the current request.
TOOLS_AND_DOCS_SYSTEM_MESSAGE_PART = """\
 You are a helpful assistant with access to the following tools. When a tool is \
required to answer the user's query, respond only with <|tool_call|> followed by a \
JSON list of tools used. If a tool does not exist in the provided list of tools, \
notify the user that you do not have the ability to fulfill the request. \

Write the response to the user's input by strictly aligning with the facts in the \
provided documents. If the information needed to answer the question is not available \
in the documents, inform the user that the question cannot be answered based on the \
available data."""


# String that a Granite 3.3 model must receive immediately after _SYSTEM_MESSAGE_START
# if there are no tools or documents in the current request and the "thinking" flag is
# set to `True`.
NO_TOOLS_AND_NO_DOCS_AND_THINKING_SYSTEM_MESSAGE_PART = f"""\
 You are a helpful AI assistant.
Respond to every user query in a comprehensive and detailed way. You can write down \
your thoughts and reasoning process before responding. In the thought process, engage \
in a comprehensive cycle of analysis, summarization, exploration, reassessment, \
reflection, backtracing, and iteration to develop well-considered thinking process. \
In the response section, based on various attempts, explorations, and reflections from \
the thoughts section, systematically present the final solution that you deem correct. \
The response should summarize the thought process. Write your thoughts between \
{COT_START}{COT_END} and write your response between \
{RESPONSE_START}{RESPONSE_END} \
for each user query."""

# String that a Granite 3.3 model must receive immediately after either
# TOOLS_AND_DOCS_SYSTEM_MESSAGE_MIDDLE  (if there are tools) or
# NO_TOOLS_AND_DOCS_SYSTEM_MESSAGE_MIDDLE (if there are no tools) in the system prompt
# if the "citations" flag is `True` and there are documents.
DOCS_AND_CITATIONS_SYSTEM_MESSAGE_PART = """ \

Use the symbols <|start_of_cite|> and <|end_of_cite|> to indicate when a fact comes \
from a document in the search result, e.g <start_of_cite> {document_id: 1}my fact \
<end_of_cite> for a fact from document 1. Afterwards, \
list all the citations with their corresponding documents in an ordered list."""

# String that a Granite 3.3 model must receive immediately after either
# TOOLS_AND_DOCS_SYSTEM_MESSAGE_MIDDLE (if there are tools and no citations) or
# NO_TOOLS_AND_DOCS_SYSTEM_MESSAGE_MIDDLE (if there are no tools or citations) or
# DOCS_AND_CITATIONS_SYSTEM_MESSAGE_PART in the system prompt
# if the "hallucinations" flag is `True` and there are documents.
# Note that a list of zero documents counts as "having documents".
DOCS_AND_HALLUCINATIONS_SYSTEM_MESSAGE_PART = """\

Finally, after the response is written, include a numbered list of sentences from the \
response with a corresponding risk value that are hallucinated and not based in the \
documents."""

# String that a Granite 3.3 model must receive immediately after _SYSTEM_MESSAGE_START
# if there are tools in the current request but there are no documents in the current
# request.
TOOLS_AND_NO_DOCS_SYSTEM_MESSAGE_PART = """\
 You are a helpful assistant with access to the following tools. When a tool is \
required to answer the user's query, respond only with <|tool_call|> followed by a \
JSON list of tools used. If a tool does not exist in the provided list of tools, \
notify the user that you do not have the ability to fulfill the request."""

MODEL_NAME = "Granite 3.3"
MODEL_HF_PATH_2B = "ibm-granite/granite-3.3-2b-instruct"



# SPDX-License-Identifier: Apache-2.0

__doc__ = f"""
{__package__} is a Python library that provides enhanced prompt creation and output
parsing for IBM Granite models
"""

# Local
# This file explicitly imports all the symbols that we export at the top level of this
# package's namespace.
from .base.types import (
    AssistantMessage,
    ChatCompletion,
    ChatCompletionResponse,
    GraniteChatCompletion,
    UserMessage,
    VLLMExtraBody,
)
from .granite3.granite32 import (
    Granite32ChatCompletion,
    Granite32InputProcessor,
    Granite32OutputProcessor,
)
from .granite3.granite33 import (
    Granite33ChatCompletion,
    Granite33InputProcessor,
    Granite33OutputProcessor,
)
from .intrinsics import IntrinsicsResultProcessor, IntrinsicsRewriter

# The contents of __all__ must be strings
__all__ = (
    obj.__name__
    for obj in (
        AssistantMessage,
        ChatCompletion,
        ChatCompletionResponse,
        UserMessage,
        Granite32InputProcessor,
        Granite32OutputProcessor,
        Granite32ChatCompletion,
        Granite33ChatCompletion,
        Granite33InputProcessor,
        Granite33OutputProcessor,
        GraniteChatCompletion,
        IntrinsicsRewriter,
        IntrinsicsResultProcessor,
        VLLMExtraBody,
    )
)



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Classes and functions that implement common aspects of input processing for all
LoRA adapters in IBM's `rag-agent-lib` library of intrinsics.
"""


# Standard
import json
import pathlib

# First Party
from granite_common import UserMessage
from granite_common.base.io import ChatCompletionRewriter
from granite_common.base.types import ChatCompletion, VLLMExtraBody
from granite_common.util import import_optional

# Local
from .constants import TOP_LOGPROBS
from .util import make_config_dict


def _needs_logprobs(transformations: list | None) -> bool:
    """
    Subroutine to check whether input processing for a model needs to specify logprobs
    in the chat completion arguments.

    :param transformations: Contents of the field by the same name in the YAML file
    :type transformations: list
    :return: ``True`` if this intrinsic produces a field for which logprobs need to be
        enabled for downstream result decoding to succeed.
    :rtype: bool
    """
    if transformations is None:
        return False
    return any(t["type"] == "likelihood" for t in transformations)


def sentence_delimiter(tag, sentence_num) -> str:
    """
    :param tag: tag string, i.e. "i" or "c"
    :param sentence_num: index of sentence
    :return: Tag string (including trailing space) that identifies the beginning of
        the indicated sentence in sentence-tagged text.
    :rtype: str
    """
    return f"<{tag}{sentence_num}> "


def mark_sentence_boundaries(
    split_strings: list[list[str]], tag_prefix: str
) -> tuple[str, int]:
    """
    Modify one or more input strings by inserting a tag in the form
    ``<[prefix][number]>``
    at the location of each sentence boundary.

    :param split_strings: Input string(s), pre-split into sentences
    :param tag_prefix: String to place before the number part of each tagged
        sentence boundary.

    :returns: List of input strings with all sentence boundaries marked.
    """
    index = 0
    result = []
    for sentences in split_strings:
        to_concat = []
        for sentence in sentences:
            to_concat.append(f"{sentence_delimiter(tag_prefix, index)}{sentence}")
            index += 1
        result.append(" ".join(to_concat))
    return result


def move_documents_to_message(
    chat_completion: ChatCompletion | dict, how: str = "string"
) -> ChatCompletion | dict:
    """
    By convention, our canned JSON requests place RAG documents in extra_body/documents.
    Some models do not accept this parameter.
    This function edits a request by putting the documents into the first turn of the
    messages.

    :param chat_completion: A chat completion request as dataclass or parsed JSON
    :param how: How to serialize the documents; supported values are "string" and "json"

    :returns: A copy of ``chat_completion`` with any documents under ``extra_body``
        moved to the first message. Returned type will be the same as the input type.
        May return original object if no edits are necessary.
    """
    if isinstance(chat_completion, ChatCompletion):
        should_return_dataclass = True
    elif isinstance(chat_completion, dict):
        should_return_dataclass = False
        chat_completion = ChatCompletion.model_validate(chat_completion)
    else:
        raise TypeError(
            f"Unexpected type '{type(chat_completion)}' for 'chat_completion' "
            f"argument. Should be ChatCompletion or dict."
        )

    if (
        chat_completion.extra_body is not None
        and chat_completion.extra_body.documents is not None
    ):
        docs_list = chat_completion.extra_body.documents

        if how == "string":
            doc_text = "\n\n".join(
                [f"[Document {d.doc_id}]\n{d.text}" for d in docs_list]
            )
            doc_message_text = (
                "You have access to the following documents:\n\n" + doc_text
            )
        elif how == "json":
            doc_message_text = json.dumps([d.model_dump() for d in docs_list])
        else:
            raise ValueError(f"Unknown document serialization method '{how}'")

        new_messages = [
            UserMessage(content=doc_message_text)
        ] + chat_completion.messages

        # Round-trip through parsed JSON so that extra_body.documents will be unset
        new_extra_body = VLLMExtraBody.model_validate(
            {
                k: v
                for k, v in chat_completion.extra_body.model_dump().items()
                if k != "documents"
            }
        )
        chat_completion = chat_completion.model_copy(
            update={"messages": new_messages, "extra_body": new_extra_body}
        )

    if should_return_dataclass:
        return chat_completion
    return chat_completion.model_dump()


class IntrinsicsRewriter(ChatCompletionRewriter):
    """General-purpose chat completion rewriter for use with the models in the
    RAG Agent Library. Reads parameters of the model's input and output formats
    from a YAML configuration file and edits the input chat completion appropriately.
    """

    config: dict
    """Parsed YAML configuration file for the target intrinsic."""

    response_format: dict
    """JSON Schema of expected response format"""

    parameters: dict
    """Additional parameters (key-value pairs) that this rewriter adds to all chat 
    completion requests."""

    extra_body_parameters: dict
    """Extended vLLM-specific parameters that go under the ``extra_body`` element of 
    the parameters field. These parameters need to be merged with any ``extra_body``
    content that is present in incoming requests."""

    instruction: str | None
    """Optional instruction template. If present, a new user message will be added with
    the indicated instruction."""

    sentence_boundaries: dict[str, str] | None
    """
    Optional sentence boundary marking specification, as a mapping from sentence 
    location (i.e. "last_message", "documents") to marker string (i.e. "c" for the
    first sentence to be marked with "<c0>").
    """

    docs_as_message: str | None
    """
    Optional specification for moving documents from ``extra_body/documents`` to a 
    user message at the beginning of the messages list. Value specifies how to serialize
    the documents into the message: "string" or "json".
    """

    def __init__(
        self,
        /,
        config_file: str | pathlib.Path | None = None,
        config_dict: dict | None = None,
        model_name: str | None = None,
    ):
        """
        :param config_file: (optional) Location of YAML configuration file
        :param config_dict: (optional) Contents of YAML configuration file that the
            caller has parsed by calling ``yaml.safe_load()``.
        :param model_name: (optional) model name to put into chat completion requests,
         or ``None`` to use default model name from the configuration
        """
        self.config = make_config_dict(config_file, config_dict)

        if self.config["parameters"] is not None and not isinstance(
            self.config["parameters"], dict
        ):
            raise TypeError(
                f"'parameters' field must be null (~ in YAML) or contain a mapping "
                f"from chat completion parameter name to value. Current value "
                f"{self.config['parameters']}"
            )

        # Split out parameters that go in extra_body
        self.parameters = self.config["parameters"] or {}
        self.extra_body_parameters = {}
        if "extra_body" in self.parameters:
            self.extra_body_parameters.update(self.parameters["extra_body"])
            del self.parameters["extra_body"]

        # Check if we're supposed to override model name
        if model_name is not None:
            self.parameters["model"] = model_name
        elif self.config["model"]:
            self.parameters["model"] = self.config["model"]

        # Compute additional parameters we need to add to every request
        if _needs_logprobs(self.config["transformations"]):
            self.parameters["logprobs"] = True
            self.parameters["top_logprobs"] = TOP_LOGPROBS
        self.instruction = self.config["instruction"]

        if self.config["response_format"] is not None:
            self.extra_body_parameters["guided_json"] = self.config["response_format"]

        self.sentence_boundaries = self.config["sentence_boundaries"]
        if self.sentence_boundaries:
            # Sentence boundary detection requires nltk
            with import_optional("nltk"):
                # Third Party
                import nltk
            self.sentence_splitter = nltk.tokenize.punkt.PunktSentenceTokenizer()
            if not isinstance(self.sentence_boundaries, dict):
                raise TypeError(
                    f"'sentence_boundaries', if present, must be a mapping from "
                    f"location (last_message/documents) to mapping prefix string. "
                    f"Received {self.sentence_boundaries}."
                )
            for k, v in self.sentence_boundaries.items():
                if k not in ("last_message", "documents"):
                    raise ValueError(
                        f"Unexpected location '{k}' in 'sentence_boundaries' field. "
                        f"Value should be 'last_message' or 'documents'."
                    )
                if not isinstance(v, str):
                    raise TypeError(
                        f"Prefix string for location '{k}' in sentence_boundaries "
                        f"field set to {v}, which is not a string."
                    )

        self.docs_as_message = self.config["docs_as_message"]
        valid_docs_as_message = ["string", "json"]
        if self.docs_as_message and self.docs_as_message not in valid_docs_as_message:
            raise ValueError(
                f"docs_as_message parameter set to '{self.docs_as_message}', which is "
                f"not one of the valid options {valid_docs_as_message}"
            )

    def _mark_sentence_boundaries(
        self, chat_completion: ChatCompletion
    ) -> ChatCompletion:
        """
        Subroutine of :func:`_transform()` that handles sentence boundary detection.

        Should be applied BEFORE adding any instruction messages to the input.

        :param chat_completion: Argument to :func:`_transform()`
        :type chat_completion: ChatCompletion
        :return: Copy of original chat completion with sentence boundaries marked in
            the last message and in documents.
        :rtype: ChatCompletion
        """
        # Mark sentence boundaries in the last message.
        if "last_message" in self.sentence_boundaries:
            messages = chat_completion.messages.copy()  # Do not modify input!
            last_message_as_sentences = list(
                self.sentence_splitter.tokenize(messages[-1].content)
            )
            rewritten_last_message_text = mark_sentence_boundaries(
                [last_message_as_sentences], self.sentence_boundaries["last_message"]
            )[0]
            messages[-1].content = rewritten_last_message_text
            chat_completion = chat_completion.model_copy(update={"messages": messages})

        # Mark sentence boundaries in documents if present
        if (
            chat_completion.extra_body.documents
            and "documents" in self.sentence_boundaries
        ):
            docs_as_sentences = [
                list(self.sentence_splitter.tokenize(d.text))
                for d in chat_completion.extra_body.documents
            ]
            # The documents input to the model consists of the original documents
            # with each sentence boundary marked with <c0>, <c1>, ... <ck-1>,
            # where `k` is the number of sentences in ALL documents.
            rewritten_docs = [
                doc.model_copy(update={"text": text})
                for doc, text in zip(
                    chat_completion.extra_body.documents,
                    mark_sentence_boundaries(
                        docs_as_sentences, self.sentence_boundaries["documents"]
                    ),
                    strict=True,
                )
            ]
            # Don't modify original input
            extra_body = chat_completion.extra_body.model_copy(
                update={"documents": rewritten_docs}
            )
            chat_completion = chat_completion.model_copy(
                update={"extra_body": extra_body}
            )
        return chat_completion

    def _transform(
        self, chat_completion: ChatCompletion, /, **kwargs
    ) -> ChatCompletion:
        edits = {}

        if self.sentence_boundaries:
            chat_completion = self._mark_sentence_boundaries(chat_completion)

        if self.docs_as_message:
            # Note that it's important for this transformation to happen after sentence
            # boundaries are inserted into the documents.
            chat_completion = move_documents_to_message(
                chat_completion, self.docs_as_message
            )

        if self.instruction is not None:
            # Generate and append new user message of instructions
            messages = chat_completion.messages.copy()  # Do not modify input!
            format_args = kwargs.copy()
            if len(messages) > 0:
                format_args["last_message"] = messages[-1].content
            try:
                instruction_str = self.instruction.format(**format_args)
            except KeyError as e:
                raise ValueError(
                    f"Missing argument for intrinsic's instruction string: {e}"
                ) from e
            messages.append(UserMessage(content=instruction_str))
            edits["messages"] = messages
        edits.update(self.parameters)

        extra_body = (
            chat_completion.extra_body.model_dump()
            if chat_completion.extra_body
            else {}
        )
        extra_body.update(self.extra_body_parameters)
        if len(extra_body) > 0:
            edits["extra_body"] = VLLMExtraBody.model_validate(extra_body)

        return chat_completion.model_copy(update=edits)



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Constants relating to of input and output processing for LoRA adapters in IBM's 
`rag-agent-lib` library of intrinsics.
"""

YAML_REQUIRED_FIELDS = [
    "model",
    "response_format",
    "transformations",
    "instruction",
    "parameters",
    "sentence_boundaries",
]
"""Fields that must be present in every intrinsic's YAML configuration file."""

YAML_JSON_FIELDS = [
    "response_format",
]
"""Fields of the YAML file that contain JSON values as strings"""

INTRINSICS_LIB_REPO_NAME = "ibm-granite/intrinsics-lib"
"""Location of the intrinsics library on Huggingface Hub"""

BASE_MODEL_TO_CANONICAL_NAME = {
    "ibm-granite/granite-3.3-8b-instruct": "granite-3.3-8b-instruct",
    "ibm-granite/granite-3.3-2b-instruct": "granite-3.3-2b-instruct",
    "granite-3.3-8b-instruct": "granite-3.3-8b-instruct",
    "granite-3.3-2b-instruct": "granite-3.3-2b-instruct",
}
"""Base model names that we accept for LoRA/aLoRA adapters in intrinsics-lib."""

TOP_LOGPROBS = 10
"""Number of logprobs we request per token when decoding logprobs."""



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Classes and functions that implement common aspects of input and output string 
processing for all Granite models.
"""

# Standard
import abc

# Third Party
import pydantic

# Local
from .types import AssistantMessage, ChatCompletion, ChatCompletionResponse


class InputProcessor(abc.ABC):
    """
    Interface for generic input processors. An input processor exposes an
    API to transform a chat completion request into a string prompt.
    """

    @abc.abstractmethod
    def transform(
        self, chat_completion: ChatCompletion, add_generation_prompt: bool = True
    ) -> str:
        """
        Convert the structured representation of the inputs to a completion request into
        the string representation of the tokens that should be sent to the model to
        implement said request.

        :param chat_completion: Structured representation of the inputs
        :param add_generation_prompt: If ``True``, the returned prompt string will
            contain a prefix of the next assistant response for use as a prompt to a
            generation request. Otherwise, the prompt will only contain the messages and
            documents in ``input``.

        :returns: String that can be passed to the model's tokenizer to create a prompt
            for generation.
        """


class OutputProcessor(abc.ABC):
    """
    Base class for generic output processors. An output processor exposes an
    API to transform model output into a structured representation of the
    information.

    This interface is very generic; see individual classes for more specific arguments.
    """

    @abc.abstractmethod
    def transform(
        self, model_output: str, chat_completion: ChatCompletion | None = None
    ) -> AssistantMessage:
        """
        Convert the model output generated into a structured representation of the
        information.

        :param model_output: String output of the a generation request, potentially
            incomplete if it was a streaming request
        :param chat_completion: The chat completion request that produced
            ``model_output``. Parameters of the request can determine how the output
            should be decoded.

        :returns: The parsed output so far, as an instance of :class:`AssistantMessage`
            possibly with model-specific extension fields.
        """


class ChatCompletionRewriter(abc.ABC):
    """
    Base class for objects that rewrite a chat completion request into another chat
    completion request.
    """

    def transform(
        self, chat_completion: ChatCompletion | str | dict, /, **kwargs
    ) -> ChatCompletion:
        """
        Rewrite a chat completion request into another chat completion request.
        Does not modify the original :class:`ChatCompletion` object.

        :param chat_completion: Original chat completion request, either as a dataclass
            or as the JSON representation of one.

        :returns: Rewritten copy of ``chat_completion``.
        """
        if isinstance(chat_completion, str):
            chat_completion = ChatCompletion.model_validate_json(chat_completion)
        if isinstance(chat_completion, dict):
            chat_completion = ChatCompletion.model_validate(chat_completion)

        if not isinstance(chat_completion, ChatCompletion):
            raise TypeError(
                f"chat_completion argument must be either a ChatCompletion "
                f"object or the JSON representation of one. Received type "
                f"'{type(chat_completion)}'."
            )
        return self._transform(chat_completion, **kwargs)

    @abc.abstractmethod
    def _transform(
        self, chat_completion: ChatCompletion, /, **kwargs
    ) -> ChatCompletion:
        """
        Subclasses must implement this internal hook to transform input requests.

        :param chat_completion: Description Original chat completion request
        :type chat_completion: ChatCompletion
        :param kwargs: Description
        :return: Description
        :rtype: ChatCompletion
        """


class ChatCompletionResultProcessor(abc.ABC):
    """
    Base class for objects that convert the raw json result of a chat completion request
    into a JSON object with model-specific postprocessing applied
    """

    def transform(
        self,
        chat_completion_response: ChatCompletionResponse | dict | pydantic.BaseModel,
        chat_completion: ChatCompletion | None = None,
    ) -> ChatCompletionResponse:
        """
        Parse the result of a chat completion.

        :param chat_completion_response: Response to a chat completion request as
            a parsed dataclass, parsed JSON value, or OpenAI dataclass
        :param chat_completion: The chat completion request that produced
            ``chat_completion_response``. Required by some implementations in order
            to decode references to part of the original request.

        :returns: Rewritten copy of ``chat_completion``.
        """
        # Convert from over-the-wire format if necessary
        if isinstance(chat_completion_response, dict):
            chat_completion_response = ChatCompletionResponse.model_validate(
                chat_completion_response
            )
        elif not isinstance(chat_completion_response, ChatCompletionResponse):
            if isinstance(chat_completion_response, pydantic.BaseModel):
                # Got another library's dataclass. Attempt to convert to our own.
                chat_completion_response = ChatCompletionResponse.model_validate(
                    chat_completion_response.model_dump()
                )
            else:
                raise TypeError(
                    f"Received unexpected type {type(chat_completion_response)=}"
                )
        return self._transform_impl(chat_completion_response, chat_completion)

    @abc.abstractmethod
    def _transform_impl(
        self,
        chat_completion_response: ChatCompletionResponse,
        chat_completion: ChatCompletion | None,
    ) -> ChatCompletionResponse:
        """
        Subclasses must override this method with an implementation of
        :func:`transform()`.
        """



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Input processing code that is specific to the Granite 3 family of models, but not
specific to a particular point release.
"""

# Standard
from collections.abc import Callable
import datetime

# First Party
from granite_common.base.io import InputProcessor
from granite_common.base.types import (
    AssistantMessage,
    SystemMessage,
    ToolResultMessage,
    UserMessage,
)
from granite_common.granite3.types import Granite3ChatCompletion


class Granite3InputProcessor(InputProcessor):
    """
    Abstract base class for Granite 3.x input processors. Contains code that is common
    across point releases.

    See the classes for the individual point release for the APIs that perform input
    transformations.
    """

    @staticmethod
    def _make_system_message_start():
        """
        :returns: String that comes at the beginning of the system message that a
        Granite 3 model must receive at the beginning of the prompt for any completion
        request that does not provide a custom system message.

        Note that the original Jinja template tends to choose weird dates from the
        future for the "Today's date" part. Instead of replicating that behavior, we
        put today's actual date in that section of the prompt. This difference probably
        doesn't matter, since none of the supervised fine tuning data exercises
        knowledge cutoffs.
        """
        return f"""\
Knowledge Cutoff Date: April 2024.
Today's Date: {datetime.datetime.now().strftime("%B %d, %Y")}.
You are Granite, developed by IBM."""

    @staticmethod
    def _split_messages(
        chat_completion: Granite3ChatCompletion,
    ) -> tuple[SystemMessage | None, list[UserMessage]]:
        """
        Separate the system message from other messages.

        :returns: Tuple of system message, if present, and remaining messages.
        """
        messages = chat_completion.messages

        # Validation code in the Inputs class should already have verified that there
        # are either zero or one system messages, and that the system message, if
        # present, occurs at position zero.
        if isinstance(messages[0], SystemMessage):
            # First message is a system message.
            return messages[0], messages[1:]
        return None, messages

    @staticmethod
    def _message_to_prompt_string(message: UserMessage | AssistantMessage) -> str:
        if isinstance(message, UserMessage):
            return (
                f"<|start_of_role|>user<|end_of_role|>{message.content}"
                f"<|end_of_text|>\n"
            )
        if isinstance(message, AssistantMessage):
            # Note that we discard any tool calls in the message, per the Jinja
            # template.
            return (
                f"<|start_of_role|>assistant<|end_of_role|>{message.content}"
                f"<|end_of_text|>\n"
            )
        if isinstance(message, ToolResultMessage):
            # Note that we discard the tool call ID, per the Jinja template.
            return (
                f"<|start_of_role|>tool<|end_of_role|>{message.content}"
                f"<|end_of_text|>\n"
            )
        raise TypeError(f"Unexpected message type {type(message)}")

    @staticmethod
    def _build_controls_record(chat_completion: Granite3ChatCompletion) -> dict | None:
        """
        Use the output control flags in ``inputs`` to build a version of the
        undocumented arbitrary JSON data regarding output controls that the Jinja
        template expected to see in the input for each chat completion request.

        :returns: A fake JSON record for "controls", or nothing of no output control
        flags were set.
        """
        if (
            not chat_completion._chat_template_kwargs()
            or not chat_completion._chat_template_kwargs().controls
        ):
            return None
        controls = chat_completion._chat_template_kwargs().controls

        result = {}
        if controls.citations:
            # The following is a guess; we have no example data for this case.
            result["citations"] = True
        if controls.hallucinations:
            # The following is a guess; we have no example data for this case.
            result["hallucinations"] = True
        if controls.length is not None:
            result["length"] = chat_completion.controls.length
        if controls.originality is not None:
            result["originality"] = chat_completion.controls.originality

        if len(result) == 0:
            return None
        return result

    @classmethod
    def _sanitize(
        cls,
        chat_completion: Granite3ChatCompletion,
        remove_special_tokens: Callable[[str], str],
        parts: list[str] | str = "all",
    ) -> Granite3ChatCompletion:
        """
        :chat_completion: Chat completion request with unsanitized inputs.
        :remove_special_tokens: Function that removes special tokens from the
            text string. Passed in subclass.
        :parts: The parts of the chat completion request to sanitize. Accepted
            values are "messages", "tools", "documents", and "all", which can be
            given individually or as part of a list. Defaults to "all".
        :returns: A new chat completion request with sanitized inputs.
        """

        # Make a copy of the chat completion object.
        chat_completion = Granite3ChatCompletion.model_validate(
            chat_completion.model_dump()
        )

        # Check given "parts" have expected values.
        sanitize_modes = ["messages", "tools", "documents", "all"]
        unsupported_parts = []
        if isinstance(parts, str):
            parts = [parts]
        for part in parts:
            if part not in sanitize_modes:
                unsupported_parts.append(part)
        if len(unsupported_parts) > 0:
            raise ValueError(
                "sanitize static method",
                "sanitize ({sanitize}) must be one of {sanitize_modes}",
                {"sanitize": unsupported_parts},
            )

        # Sanitize based on the given parts.
        if ("messages" in parts or "all" in parts) and chat_completion.messages:
            for message in chat_completion.messages:
                message.content = remove_special_tokens(message.content)
        if ("tools" in parts or "all" in parts) and chat_completion.tools:
            for tool in chat_completion.tools:
                tool.name = remove_special_tokens(tool.name)
                if tool.description:
                    tool.description = remove_special_tokens(tool.description)
                if tool.parameters:
                    new_params = {}
                    for k, v in tool.parameters.items():
                        kk = remove_special_tokens(k)
                        vv = remove_special_tokens(v)
                        if len(kk) > 0:
                            new_params[kk] = vv
                    tool.parameters = new_params
        if ("documents" in parts or "all" in parts) and chat_completion._documents():
            for document in chat_completion._documents():
                if document.doc_id and isinstance(document.doc_id, str):
                    document.doc_id = remove_special_tokens(document.doc_id)
                document.text = remove_special_tokens(document.text)

        return chat_completion



# SPDX-License-Identifier: Apache-2.0

__doc__ = """
Tests of code under ``granite_common.rag_agent_lib``
"""

# Standard
import copy
import json
import os
import pathlib

# Third Party
import huggingface_hub
import openai
import pytest
import requests
import yaml

# First Party
from granite_common import ChatCompletion, IntrinsicsRewriter
from granite_common.base.types import ChatCompletionResponse
from granite_common.intrinsics import json_util, util
from granite_common.intrinsics.constants import RAG_INTRINSICS_LIB_REPO_NAME
from granite_common.intrinsics.output import IntrinsicsResultProcessor
import granite_common.util


def _read_file(name):
    with open(name, encoding="utf-8") as f:
        return f.read()


_TEST_DATA_DIR = pathlib.Path(os.path.dirname(__file__)) / "testdata"

# Base model to use for testing; should be small enough to run in memory on the CI
# server.
_BASE_MODEL = "granite-3.3-2b-instruct"

_INPUT_JSON_DIR = _TEST_DATA_DIR / "input_json"
_INPUT_YAML_DIR = _TEST_DATA_DIR / "input_yaml"
_INPUT_ARGS_DIR = _TEST_DATA_DIR / "input_args"


# Combinations of YAML and JSON files that go together.
_YAML_JSON_COMBOS = {
    # Short name => YAML file, JSON file, model file, arguments file, is aLoRA
    "answerability_simple": (
        _INPUT_YAML_DIR / "answerability.yaml",
        _INPUT_JSON_DIR / "simple.json",
        "answerability",
        None,
        False,
    ),
    "answerability_extra_params": (
        _INPUT_YAML_DIR / "answerability.yaml",
        _INPUT_JSON_DIR / "extra_params.json",
        None,
        None,
        False,
    ),
    "answerability_answerable": (
        _INPUT_YAML_DIR / "answerability.yaml",
        _INPUT_JSON_DIR / "answerable.json",
        "answerability",
        None,
        False,
    ),
    "answerability_answerable_alora": (
        _INPUT_YAML_DIR / "answerability.yaml",
        _INPUT_JSON_DIR / "answerable.json",
        "answerability",
        None,
        True,
    ),
    "answerability_unanswerable": (
        _INPUT_YAML_DIR / "answerability.yaml",
        _INPUT_JSON_DIR / "unanswerable.json",
        "answerability",
        None,
        False,
    ),
    "answerability_unanswerable_alora": (
        _INPUT_YAML_DIR / "answerability.yaml",
        _INPUT_JSON_DIR / "unanswerable.json",
        "answerability",
        None,
        True,
    ),
    "instruction": (
        _INPUT_YAML_DIR / "instruction.yaml",
        _INPUT_JSON_DIR / "instruction.json",
        None,  # Fake config, no model
        _INPUT_ARGS_DIR / "instruction.json",
        False,  # No model -> no aLoRA
    ),
    "hallucination_detection": (
        _INPUT_YAML_DIR / "hallucination_detection.yaml",
        _INPUT_JSON_DIR / "hallucination_detection.json",
        "hallucination_detection",
        None,
        False,
    ),
    "query_rewrite": (
        _INPUT_YAML_DIR / "query_rewrite.yaml",
        _INPUT_JSON_DIR / "query_rewrite.json",
        "query_rewrite",
        None,
        False,
    ),
    "requirement_check": (
        _INPUT_YAML_DIR / "requirement_check.yaml",
        _INPUT_JSON_DIR / "requirement_check.json",
        "requirement_check",
        _INPUT_ARGS_DIR / "requirement_check.json",
        False,
    ),
    "requirement_check_alora": (
        _INPUT_YAML_DIR / "requirement_check.yaml",
        _INPUT_JSON_DIR / "requirement_check.json",
        "requirement_check",
        _INPUT_ARGS_DIR / "requirement_check.json",
        True,
    ),
    "uncertainty": (
        _INPUT_YAML_DIR / "uncertainty.yaml",
        _INPUT_JSON_DIR / "uncertainty.json",
        "uncertainty",
        None,
        False,
    ),
    "uncertainty_alora": (
        _INPUT_YAML_DIR / "uncertainty.yaml",
        _INPUT_JSON_DIR / "uncertainty.json",
        "uncertainty",
        None,
        True,
    ),
    "context_relevance": (
        _INPUT_YAML_DIR / "context_relevance.yaml",
        _INPUT_JSON_DIR / "context_relevance.json",
        "context_relevance",
        _INPUT_ARGS_DIR / "context_relevance.json",
        False,
    ),
    "answer_relevance_classifier": (
        _INPUT_YAML_DIR / "answer_relevance_classifier.yaml",
        _INPUT_JSON_DIR / "answer_relevance_classifier.json",
        "answer_relevance_classifier",
        None,
        False,
    ),
    "answer_relevance_classifier_alora": (
        _INPUT_YAML_DIR / "answer_relevance_classifier.yaml",
        _INPUT_JSON_DIR / "answer_relevance_classifier.json",
        "answer_relevance_classifier",
        None,
        True,
    ),
    "answer_relevance_rewriter": (
        _INPUT_YAML_DIR / "answer_relevance_rewriter.yaml",
        _INPUT_JSON_DIR / "answer_relevance_rewriter.json",
        "answer_relevance_rewriter",
        _INPUT_ARGS_DIR / "answer_relevance_rewriter.json",
        False,
    ),
    "answer_relevance_rewriter_alora": (
        _INPUT_YAML_DIR / "answer_relevance_rewriter.yaml",
        _INPUT_JSON_DIR / "answer_relevance_rewriter.json",
        "answer_relevance_rewriter",
        _INPUT_ARGS_DIR / "answer_relevance_rewriter.json",
        True,
    ),
    "citations": (
        _INPUT_YAML_DIR / "citations.yaml",
        _INPUT_JSON_DIR / "citations.json",
        "citations",
        None,
        False,
    ),
    "citations_alora": (
        _INPUT_YAML_DIR / "citations.yaml",
        _INPUT_JSON_DIR / "citations.json",
        "citations",
        None,
        True,
    ),
}


# All combinations of input and model
_YAML_JSON_COMBOS_WITH_MODEL = {
    k: v for k, v in _YAML_JSON_COMBOS.items() if v[2] is not None
}

# All combinations of input and model that are not aLoRA models (includes no model)
_YAML_JSON_COMBOS_NO_ALORA = {
    k: v[:4] for k, v in _YAML_JSON_COMBOS.items() if not v[4]
}


@pytest.fixture(name="yaml_json_combo", scope="module", params=_YAML_JSON_COMBOS)
def _yaml_json_combo(request: pytest.FixtureRequest) -> tuple[str, str, str, str]:
    """Pytest fixture that allows us to run a given test case repeatedly with multiple
    different combinations of IO configuration and chat completion request.

    Uses the files in ``testdata/input_json`` and ``testdata/input_yaml``.

    Returns tuple of short name, YAML file, JSON file, model directory, and
    arguments file.
    """
    return (request.param,) + _YAML_JSON_COMBOS[request.param]


@pytest.fixture(
    name="yaml_json_combo_no_alora", scope="module", params=_YAML_JSON_COMBOS_NO_ALORA
)
def _yaml_json_combo_no_alora(
    request: pytest.FixtureRequest,
) -> tuple[str, str, str, str]:
    """Pytest fixture that allows us to run a given test case repeatedly with multiple
    different combinations of IO configuration and chat completion request. Ignores
    model configs that use the aLoRA variant of the model.

    Uses the files in ``testdata/input_json`` and ``testdata/input_yaml``.

    Returns tuple of short name, YAML file, JSON file, model directory, and
    arguments file.
    """
    return (request.param,) + _YAML_JSON_COMBOS_NO_ALORA[request.param]


@pytest.fixture(
    name="yaml_json_combo_with_model",
    scope="module",
    params=_YAML_JSON_COMBOS_WITH_MODEL,
)
def _yaml_json_combo_with_model(request: pytest.FixtureRequest) -> tuple[str, str, str]:
    """Version of :func:`_yaml_json_combo()` fixture with only the inputs that have
    models. Includes an additional flag for whether the model is LoRA or aLoRA
    """
    return (request.param,) + _YAML_JSON_COMBOS_WITH_MODEL[request.param]


def test_no_orphan_files():
    """Check whether there are input files that aren't used by any test."""
    used_json_files = set(t[1].name for t in _YAML_JSON_COMBOS.values())
    all_json_files = os.listdir(_INPUT_JSON_DIR)
    used_yaml_files = set(t[0].name for t in _YAML_JSON_COMBOS.values())
    all_yaml_files = os.listdir(_INPUT_YAML_DIR)

    for f in all_json_files:
        if f not in used_json_files:
            raise ValueError(
                f"JSON File '{f}' not used. Files are {all_json_files}; "
                f"Used files are {list(used_json_files)}"
            )
    for f in all_yaml_files:
        if f not in used_yaml_files:
            raise ValueError(
                f"YAML File '{f}' not used. Files are {all_yaml_files}; "
                f"Used files are {list(used_yaml_files)}"
            )


def test_read_yaml():
    """Sanity check to verify that reading a model's YAML file from disk works."""
    # Read from local disk
    with open(_INPUT_YAML_DIR / "answerability.yaml", encoding="utf8") as file:
        data = yaml.safe_load(file)
    assert data["model"] is None

    original_data = copy.deepcopy(data)

    # Instantiate directly from dictionary
    IntrinsicsRewriter(config_dict=data)

    # Data shouldn't be modified
    assert original_data == data

    # Manually run through the make_config_dict() function, because apparently users
    # will try to do this.
    data2 = util.make_config_dict(_INPUT_YAML_DIR / "answerability.yaml")
    IntrinsicsRewriter(config_dict=data2)

    # Read from local disk
    IntrinsicsRewriter(config_file=_INPUT_YAML_DIR / "answerability.yaml")

    # Read from Hugging Face hub.
    # Requires "hf auth login" with read token while repo is private.
    path_suffix = "answerability/lora/granite-3.3-2b-instruct/io.yaml"
    try:
        local_path = huggingface_hub.snapshot_download(
            repo_id=RAG_INTRINSICS_LIB_REPO_NAME,
            allow_patterns=path_suffix,
        )
    except requests.exceptions.HTTPError:
        pytest.xfail("Downloads fail on CI server because repo is private")
    IntrinsicsRewriter(config_file=f"{local_path}/{path_suffix}")


_CANNED_INPUT_EXPECTED_DIR = _TEST_DATA_DIR / "test_canned_input"


def test_canned_input(yaml_json_combo_no_alora):
    """
    Verify that a given combination of chat completion and rewriting config produces
    the expected output
    """
    short_name, yaml_file, json_file, _, args_file = yaml_json_combo_no_alora
    if args_file:
        with open(args_file, encoding="utf8") as f:
            transform_kwargs = json.load(f)
    else:
        transform_kwargs = {}

    # Temporary: Use a YAML file from local disk
    rewriter = IntrinsicsRewriter(config_file=yaml_file)

    json_data = _read_file(json_file)
    before = ChatCompletion.model_validate_json(json_data)
    after = rewriter.transform(before, **transform_kwargs)
    after_json = after.model_dump_json(indent=2)

    expected_file = _CANNED_INPUT_EXPECTED_DIR / f"{short_name}.json"
    with open(expected_file, encoding="utf-8") as f:
        expected_json = f.read()

    print(f"{after_json=}")
    assert after_json == expected_json


@pytest.mark.block_network
def test_openai_compat(yaml_json_combo_no_alora: str):
    """
    Verify that the dataclasses for intrinsics chat completions can be directly passed
    to the OpenAI Python API without raising parsing errors.
    """

    _, yaml_file, json_file, _, args_file = yaml_json_combo_no_alora
    if args_file:
        with open(args_file, encoding="utf8") as f:
            transform_kwargs = json.load(f)
    else:
        transform_kwargs = {}

    # Temporary: Use a YAML file from local disk
    rewriter = IntrinsicsRewriter(config_file=yaml_file)
    json_data = _read_file(json_file)
    before = ChatCompletion.model_validate_json(json_data)
    after = rewriter.transform(before, **transform_kwargs)

    # Create a fake connection to the API so we can use its request validation code.
    # Note that network access is blocked for this test case.
    openai_base_url = "http://localhost:98765/not/a/valid/url"
    openai_api_key = "not_a_valid_api_key"
    client = openai.OpenAI(base_url=openai_base_url, api_key=openai_api_key)

    # OpenAI requires a model name
    before.model = "dummy_model_name"
    after.model = "dummy_model_name"

    # The client should get all the way through validation and fail to connect
    with pytest.raises(openai.APIConnectionError):
        client.chat.completions.create(**(before.model_dump()))
    with pytest.raises(openai.APIConnectionError):
        client.chat.completions.create(**(after.model_dump()))


# Combinations of YAML and canned output files that go together.
# Canned output is in test_canned_output/model_output/<short name>.json
_YAML_OUTPUT_COMBOS = {
    # Short name => YAML file
    "answerability_answerable": _INPUT_YAML_DIR / "answerability.yaml",
    "answerability_unanswerable": _INPUT_YAML_DIR / "answerability.yaml",
    "query_rewrite": _INPUT_YAML_DIR / "query_rewrite.yaml",
    "context_relevance": _INPUT_YAML_DIR / "context_relevance.yaml",
    "hallucination_detection": _INPUT_YAML_DIR / "hallucination_detection.yaml",
    "citations": _INPUT_YAML_DIR / "citations.yaml",
    "requirement_check": _INPUT_YAML_DIR / "requirement_check.yaml",
    "answer_relevance_classifier": _INPUT_YAML_DIR / "answer_relevance_classifier.yaml",
    "answer_relevance_rewriter": _INPUT_YAML_DIR / "answer_relevance_rewriter.yaml",
}


_CANNED_OUTPUT_MODEL_INPUT_DIR = _TEST_DATA_DIR / "test_canned_output/model_input"
_CANNED_OUTPUT_MODEL_OUTPUT_DIR = _TEST_DATA_DIR / "test_canned_output/model_output"
_CANNED_OUTPUT_EXPECTED_DIR = _TEST_DATA_DIR / "test_canned_output/expected_result"


@pytest.fixture(name="yaml_output_combo", scope="module", params=_YAML_OUTPUT_COMBOS)
def _yaml_output_combo(request: pytest.FixtureRequest) -> tuple[str, str]:
    """Pytest fixture that iterates over the various inputs to
    :func:`test_canned_output()`

    :returns: Tuple of:
     * short name of test case
     * location of YAML file
     * location of model input file
     * location of model raw output file
     * location of expected file
    """
    return (
        request.param,
        _YAML_OUTPUT_COMBOS[request.param],
        _CANNED_OUTPUT_MODEL_INPUT_DIR / f"{request.param}.json",
        _CANNED_OUTPUT_MODEL_OUTPUT_DIR / f"{request.param}.json",
        _CANNED_OUTPUT_EXPECTED_DIR / f"{request.param}.json",
    )


def test_canned_output(yaml_output_combo):
    """
    Verify that the output processing for each model works on previous model outputs
    read from disk. Model outputs are stored in OpenAI format.

    :param yaml_output_combo: Fixture containing pairs of short name, IO YAML file
    """
    _, yaml_file, input_file, output_file, expected_file = yaml_output_combo

    processor = IntrinsicsResultProcessor(config_file=yaml_file)
    with open(input_file, encoding="utf-8") as f:
        model_input = ChatCompletion.model_validate_json(f.read())
    with open(output_file, encoding="utf-8") as f:
        model_output = ChatCompletionResponse.model_validate_json(f.read())

    transformed = processor.transform(model_output, model_input)

    # Pull this string out of the debugger to update expected file
    transformed_str = transformed.model_dump_json(indent=4)

    with open(expected_file, encoding="utf-8") as f:
        expected = ChatCompletionResponse.model_validate_json(f.read())
    expected_str = expected.model_dump_json(indent=4)

    # Do an approximate comparison of numeric values.
    # Can't use pytest.approx() because of lists and floats encoded as strings
    transformed_json = _round_floats(json.loads(transformed_str))
    expected_json = _round_floats(json.loads(expected_str))

    assert transformed_json == expected_json


_REPARSE_JSON_DIR = _TEST_DATA_DIR / "test_reparse_json"
_REPARSE_JSON_FILES = [
    name for name in os.listdir(_REPARSE_JSON_DIR) if name.endswith(".json")
]


@pytest.fixture(name="reparse_json_file", scope="module", params=_REPARSE_JSON_FILES)
def _reparse_json_file(request: pytest.FixtureRequest) -> tuple[str, str, str]:
    """Pytest fixture that returns each file in _REPARSE_JSON_DIR in turn"""
    return request.param


def test_reparse_json(reparse_json_file):
    """Ensure that we can reparse JSON data to find position information for
    literals."""
    json_file = _REPARSE_JSON_DIR / reparse_json_file
    json_str = _read_file(json_file)

    parsed_json = json.loads(json_str)
    reparsed_json = json_util.reparse_json_with_offsets(json_str)

    assert json_util.scalar_paths(parsed_json) == json_util.scalar_paths(reparsed_json)


def _round_floats(json_data, num_digits: int = 2):
    """Round all floating-point numbers in a JSON value to facilitate comparisons.

    :param json_data: Arbitrary JSON data.
    :param num_digits: How many decimal points to round to

    :returns: Copy of the input with all floats rounded
    """
    result = copy.deepcopy(json_data)
    for path in json_util.scalar_paths(result):
        value = json_util.fetch_path(result, path)
        if isinstance(value, float):
            json_util.replace_path(result, path, round(value, num_digits))
        elif isinstance(value, str):
            # Test for floating-point number encoded as a string.
            # In Python this test is supposed to use exceptions as control flow.
            try:
                str_as_float = float(value)
                json_util.replace_path(result, path, round(str_as_float, num_digits))
            except ValueError:
                # flow through
                pass

            # Test for JSON object or array encoded as a string
            if value[0] in ("{", "["):
                try:
                    str_as_json = json.loads(value)
                    rounded_json = _round_floats(str_as_json, num_digits)
                    rounded_json_as_str = json.dumps(rounded_json)
                    json_util.replace_path(result, path, rounded_json_as_str)
                except json.JSONDecodeError:
                    # flow through
                    pass
    return result


def test_run_transformers(yaml_json_combo_with_model):
    """
    Run the target model end-to-end on transformers.
    """
    short_name, yaml_file, input_file, model_name, args_file, alora = (
        yaml_json_combo_with_model
    )
    if args_file:
        with open(args_file, encoding="utf8") as f:
            transform_kwargs = json.load(f)
    else:
        transform_kwargs = {}

    # Load input request
    with open(input_file, encoding="utf-8") as f:
        model_input = ChatCompletion.model_validate_json(f.read())

    # Download files from Hugging Face Hub
    try:
        lora_dir = util.obtain_lora(model_name, _BASE_MODEL, alora=alora)
    except requests.exceptions.HTTPError:
        pytest.xfail("Downloads fail on CI server because repo is private")

    # Load IO config YAML for this model
    io_yaml_path = lora_dir / "io.yaml"
    if not os.path.exists(io_yaml_path):
        # Use local files until proper configs are up on Hugging Face
        io_yaml_path = yaml_file
    rewriter = IntrinsicsRewriter(config_file=io_yaml_path)
    result_processor = IntrinsicsResultProcessor(config_file=io_yaml_path)

    # Prepare inputs for inference
    transformed_input = rewriter.transform(model_input, **transform_kwargs)

    # Run the model using Hugging Face APIs
    model, tokenizer = granite_common.util.load_transformers_lora(lora_dir)
    generate_input, other_input = (
        granite_common.util.chat_completion_request_to_transformers_inputs(
            transformed_input.model_dump(), tokenizer, model
        )
    )
    responses = granite_common.util.generate_with_transformers(
        tokenizer, model, generate_input, other_input
    )

    # Output processing
    transformed_responses = result_processor.transform(responses, transformed_input)

    # Pull this string out of the debugger to create a fresh expected file.
    transformed_str = transformed_responses.model_dump_json(indent=4)
    print(transformed_str)

    with open(
        _TEST_DATA_DIR / f"test_run_transformers/{short_name}.json", encoding="utf-8"
    ) as f:
        expected = ChatCompletionResponse.model_validate_json(f.read())
    # expected_str = expected.model_dump_json(indent=4)

    # Correct for floating point rounding.
    # Can't use pytest.approx() because of lists
    transformed_json = _round_floats(
        json_util.parse_inline_json(transformed_responses.model_dump())
    )
    expected_json = _round_floats(json_util.parse_inline_json(expected.model_dump()))
    assert transformed_json == expected_json
